@@ -7,6 +7,7 @@
   PrettyPrec printing class and instances for CoreHW
 -}
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE FlexibleInstances #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
@@ -36,11 +37,20 @@ import Control.Monad.Identity
 import Data.List.Extra                  ((<:>))
 import qualified Data.Text              as T
 import Data.Maybe                       (fromMaybe)
+#if MIN_VERSION_prettyprinter(1,7,0)
+import Prettyprinter
+import Prettyprinter.Internal
+#else
 import Data.Text.Prettyprint.Doc
 import Data.Text.Prettyprint.Doc.Internal
+#endif
 import GHC.Show                         (showMultiLineString)
 import Numeric                          (fromRat)
+#if MIN_VERSION_ghc(9,0,0)
+import qualified GHC.Utils.Outputable   as GHC
+#else
 import qualified Outputable             as GHC
+#endif
 import System.Environment               (lookupEnv)
 import System.IO.Unsafe                 (unsafePerformIO)
 import Text.Read                        (readMaybe)
@@ -79,12 +89,16 @@ data PrettyOptions = PrettyOptions
   -- ^ whether to display type information
   , displayQualifiers :: Bool
   -- ^ whether to display module qualifiers
+  , displayTicks      :: Bool
+  -- ^ whether to display ticks
   }
+
 instance Default PrettyOptions where
   def = PrettyOptions
     { displayUniques    = unsafeLookupEnvBool "CLASH_PPR_UNIQUES" True
     , displayTypes      = unsafeLookupEnvBool "CLASH_PPR_TYPES" True
     , displayQualifiers = unsafeLookupEnvBool "CLASH_PPR_QUALIFIERS" True
+    , displayTicks      = unsafeLookupEnvBool "CLASH_PPR_TICKS" True
     }
 
 -- | Annotations carried on pretty-printed code.
@@ -96,7 +110,7 @@ data ClashAnnotation
   deriving Eq
 
 -- | Specific places in the program syntax.
-data SyntaxElement = Keyword | LitS | Type | Unique | Qualifier
+data SyntaxElement = Keyword | LitS | Type | Unique | Qualifier | Ticky
   deriving (Eq, Show)
 
 -- | Clash's specialized @Doc@ type holds metadata of type @ClashAnnotation@.
@@ -125,6 +139,7 @@ class PrettyPrec p where
           if not (displayTypes opts)      && ann == AnnSyntax Type
           || not (displayUniques opts)    && ann == AnnSyntax Unique
           || not (displayQualifiers opts) && ann == AnnSyntax Qualifier
+          || not (displayTicks opts)      && ann == AnnSyntax Ticky
             then Empty
             else Annotated ann (hide d')
         d -> d
@@ -251,7 +266,7 @@ instance PrettyPrec Term where
     Tick t e'       -> do
       tDoc <- pprPrec prec t
       eDoc <- pprPrec prec e'
-      return (tDoc <> line' <> eDoc)
+      return (annotate (AnnSyntax Ticky) (tDoc <> line') <> eDoc)
 
 instance PrettyPrec TickInfo where
   pprPrec prec (SrcSpan sp)   = pprPrec prec sp
