@@ -57,7 +57,13 @@ module Clash.Core.Subst
 where
 
 import           Data.Coerce               (coerce)
+
+#if MIN_VERSION_prettyprinter(1,7,0)
+import           Prettyprinter
+#else
 import           Data.Text.Prettyprint.Doc
+#endif
+
 import qualified Data.List                 as List
 import qualified Data.List.Extra           as List
 import           Data.Ord                  (comparing)
@@ -420,7 +426,7 @@ substInExistentials is exts subst@(typeVar, _type) =
       -- it is as we would replace global variables in them.
       take (i+1) exts ++ substGlobalsInExistentials is (drop (i+1) exts) [subst]
 
--- | This checks if the substitution satisfies the invariant from 'TvSbust's
+-- | This checks if the substitution satisfies the invariant from 'TvSubst's
 -- Note [The substitution invariant].
 checkValidSubst
   :: HasCallStack
