@@ -12,6 +12,7 @@
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
+{-# LANGUAGE QuasiQuotes #-}
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TemplateHaskell #-}
 
@@ -23,6 +24,7 @@ import           Control.Lens
 import           Control.Monad                   (forM)
 import           Control.Monad.State             (State, StateT (..), lift)
 import           Data.Bool                       (bool)
+import           Data.Coerce                     (coerce)
 import           Data.Foldable                   (foldrM)
 import           Data.Hashable                   (Hashable (..))
 import qualified Data.IntMap                     as IntMap
@@ -36,7 +38,13 @@ import           Data.Semigroup.Monad
 import qualified Data.Text
 import           Data.Text.Lazy                  (Text)
 import qualified Data.Text.Lazy                  as Text
+
+#if MIN_VERSION_prettyprinter(1,7,0)
+import qualified Prettyprinter                   as PP
+#else
 import qualified Data.Text.Prettyprint.Doc       as PP
+#endif
+
 import           Data.Text.Prettyprint.Doc.Extra
 import           System.FilePath                 (replaceBaseName, takeBaseName,
                                                   takeFileName, (<.>))
@@ -44,25 +52,24 @@ import           Text.Printf
 import           Text.Read                       (readEither)
 import           Text.Trifecta.Result            hiding (Err)
 
-import           Clash.Backend                   (Backend (..), Usage (..), mkUniqueIdentifier)
-import qualified Clash.Backend                   as Backend
+import           Clash.Backend
+  (Backend (..), Usage (..), AggressiveXOptBB(..))
 import           Clash.Netlist.BlackBox.Parser
 import           Clash.Netlist.BlackBox.Types
-import           Clash.Netlist.Id                (IdType (..))
-import           Clash.Netlist.Types             (BlackBoxContext (..),
-                                                  Expr (..), HWType (..),
-                                                  Identifier, Literal (..),
-                                                  Modifier (..),
-                                                  Declaration(BlackBoxD))
+import           Clash.Netlist.Types
+  (BlackBoxContext (..), Expr (..), HWType (..), Literal (..), Modifier (..),
+   Declaration(BlackBoxD))
+import qualified Clash.Netlist.Id                as Id
 import qualified Clash.Netlist.Types             as N
 import           Clash.Netlist.Util              (typeSize, isVoid, stripVoid)
 import           Clash.Signal.Internal
   (ResetKind(..), ResetPolarity(..), InitBehavior(..))
 import           Clash.Util
+import qualified Clash.Util.Interpolate          as I
 
 inputHole :: Element -> Maybe Int
 inputHole = \case
-  Arg _ n       -> pure n
+  Arg n         -> pure n
   Lit n         -> pure n
   Const n       -> pure n
   Name n        -> pure n
@@ -144,12 +151,11 @@ extractLiterals = map (\case (e,_,_) -> e)
 -- counter for every newly encountered symbol.
 setSym
   :: forall m
-   . Monad m
-  => (IdType -> Identifier -> m Identifier)
-  -> BlackBoxContext
+   . Id.IdentifierSetMonad m
+  => BlackBoxContext
   -> BlackBoxTemplate
   -> m (BlackBoxTemplate,[N.Declaration])
-setSym mkUniqueIdentifierM bbCtx l = do
+setSym bbCtx l = do
     (a,(_,decls)) <- runStateT (mapM setSym' l) (IntMap.empty,IntMap.empty)
     return (a,concatMap snd (IntMap.elems decls))
   where
@@ -157,33 +163,34 @@ setSym mkUniqueIdentifierM bbCtx l = do
 
     setSym'
       :: Element
-      -> StateT ( IntMap.IntMap Identifier
-                , IntMap.IntMap (Identifier,[N.Declaration]))
+      -> StateT ( IntMap.IntMap N.IdentifierText
+                , IntMap.IntMap (N.IdentifierText, [N.Declaration]))
                 m
                 Element
     setSym' e = case e of
       ToVar nm i | i < length (bbInputs bbCtx) -> case bbInputs bbCtx !! i of
-        (Identifier nm' Nothing,_,_) ->
-          return (ToVar [Text (Text.fromStrict nm')] i)
+        (Identifier nm0 Nothing,_,_) ->
+          return (ToVar [Text (Id.toLazyText nm0)] i)
 
         (e',hwTy,_) -> do
           varM <- IntMap.lookup i <$> use _2
           case varM of
             Nothing -> do
-              nm' <- lift (mkUniqueIdentifierM Extended (Text.toStrict (concatT (Text "c$":nm))))
+              nm' <- lift (Id.make (Text.toStrict (concatT (Text "c$":nm))))
               let decls = case typeSize hwTy of
                     0 -> []
                     _ -> [N.NetDecl Nothing nm' hwTy
                          ,N.Assignment nm' e'
                          ]
-              _2 %= (IntMap.insert i (nm',decls))
+              _2 %= (IntMap.insert i (Id.toText nm',decls))
+              return (ToVar [Text (Id.toLazyText nm')] i)
+            Just (nm',_) ->
               return (ToVar [Text (Text.fromStrict nm')] i)
-            Just (nm',_) -> return (ToVar [Text (Text.fromStrict nm')] i)
       Sym _ i -> do
         symM <- IntMap.lookup i <$> use _1
         case symM of
           Nothing -> do
-            t <- lift (mkUniqueIdentifierM Extended "c$n")
+            t <- Id.toText <$> lift (Id.make "c$n")
             _1 %= (IntMap.insert i t)
             return (Sym (Text.fromStrict t) i)
           Just t -> return (Sym (Text.fromStrict t) i)
@@ -191,7 +198,7 @@ setSym mkUniqueIdentifierM bbCtx l = do
         symM <- IntMap.lookup i <$> use _1
         case symM of
           Nothing -> do
-            t' <- lift (mkUniqueIdentifierM Basic (Text.toStrict (concatT t)))
+            t' <- Id.toText <$> lift (Id.makeBasic (Text.toStrict (concatT t)))
             _1 %= (IntMap.insert i t')
             return (GenSym [Text (Text.fromStrict t')] i)
           Just _ ->
@@ -223,12 +230,12 @@ setSym mkUniqueIdentifierM bbCtx l = do
               error $ $(curLoc) ++  "Could not convert ~LIT[" ++ show i ++ "]"
                    ++ " to string:" ++ msg ++ "\n\nError occured while "
                    ++ "processing blackbox for " ++ bbnm
-        Result _ | Identifier t _ <- fst (bbResult bbCtx) -> Text.fromStrict t
-        CompName -> Text.fromStrict (bbCompName bbCtx)
+        Result | [(Identifier t _, _)] <- bbResults bbCtx -> Id.toLazyText t
+        CompName -> Id.toLazyText (bbCompName bbCtx)
         CtxName ->
           case bbCtxName bbCtx of
             Just nm -> Text.fromStrict nm
-            _ | Identifier t _ <- fst (bbResult bbCtx) -> Text.fromStrict t
+            _ | [(Identifier t _, _)] <- bbResults bbCtx -> Id.toLazyText t
             _ -> error $ $(curLoc) ++ "Internal error when processing blackbox "
                       ++ "for " ++ bbnm
         _ -> error $ $(curLoc) ++ "Unexpected element in GENSYM when processing "
@@ -331,7 +338,7 @@ renderElem b (Component (Decl n subN (l:ls))) = do
                     , show (subN +1 ), " got only ", show (length (fromJust func0)) ]
       func1 = indexNote' errr subN <$> func0
       Just (templ0,_,libs,imps,inc,pCtx) = func1
-      b' = pCtx { bbResult = (o,oTy), bbInputs = bbInputs pCtx ++ is }
+      b' = pCtx { bbResults = [(o,oTy)], bbInputs = bbInputs pCtx ++ is }
       layoutOptions = LayoutOptions (AvailablePerLine 120 0.4)
       render = N.BBTemplate . parseFail . renderLazy . layoutPretty layoutOptions
 
@@ -340,7 +347,7 @@ renderElem b (Component (Decl n subN (l:ls))) = do
       Left t ->
         return t
       Right (nm0,ds) -> do
-        nm1 <- mkUniqueIdentifier Basic nm0
+        nm1 <- Id.next nm0
         block <- getMon (blockDecl nm1 ds)
         return (render block)
 
@@ -349,13 +356,13 @@ renderElem b (Component (Decl n subN (l:ls))) = do
       N.BBFunction {} ->
         return templ1
       N.BBTemplate templ2 -> do
-        (templ3, templDecls) <- setSym Backend.mkUniqueIdentifier b' templ2
+        (templ3, templDecls) <- setSym b' templ2
         case templDecls of
           [] ->
             return (N.BBTemplate templ3)
           _ -> do
-            nm1 <- Backend.mkUniqueIdentifier Basic "bb"
-            nm2 <- Backend.mkUniqueIdentifier Basic "bb"
+            nm1 <- Id.toText <$> Id.makeBasic "bb"
+            nm2 <- Id.makeBasic "bb"
             let bbD = BlackBoxD nm1 libs imps inc (N.BBTemplate templ3) b'
             block <- getMon (blockDecl nm2 (templDecls ++ [bbD]))
             return (render block)
@@ -374,7 +381,7 @@ renderElem b (Component (Decl n subN (l:ls))) = do
 renderElem b (SigD e m) = do
   e' <- Text.concat <$> mapM (fmap ($ 0) . renderElem b) e
   let ty = case m of
-             Nothing -> snd $ bbResult b
+             Nothing -> snd $ bbResult "~SIGD" b
              Just n  -> let (_,ty',_) = bbInputs b !! n
                         in  ty'
   t  <- getMon (hdlSig e' ty)
@@ -393,9 +400,11 @@ renderElem b (Tag n) = do
   case stripVoid ty of
     KnownDomain dom _ _ _ _ _ ->
       return (const (Text.pack (Data.Text.unpack dom)))
+    Clock dom ->
+      return (const (Text.pack (Data.Text.unpack dom)))
     Reset dom ->
       return (const (Text.pack (Data.Text.unpack dom)))
-    Clock dom ->
+    Enable dom ->
       return (const (Text.pack (Data.Text.unpack dom)))
     _ ->
       error $ $(curLoc) ++ "Tag: Expected `KnownDomain` or `KnownConfiguration`, not: " ++ show ty
@@ -404,10 +413,12 @@ renderElem b (Tag n) = do
 renderElem b (IF c t f) = do
   iw <- iwWidth
   syn <- hdlSyn
-  let c' = check iw syn c
+  xOpt <- aggressiveXOptBB
+  let c' = check (coerce xOpt) iw syn c
   if c' > 0 then renderTemplate b t else renderTemplate b f
   where
-    check iw syn c' = case c' of
+    check :: Bool -> Int -> HdlSyn -> Element -> Int
+    check xOpt iw syn c' = case c' of
       (Size e)   -> typeSize (lineToType b [e])
       (Length e) -> case lineToType b [e] of
                        (Vector n _)             -> n
@@ -446,15 +457,31 @@ renderElem b (IF c t f) = do
                       BlackBoxE {} -> 1
                       _            -> 0
 
+      (IsUndefined n) ->
+        let (e, _, _) = bbInputs b !! n in
+        case (xOpt, e) of
+          (True, BlackBoxE _ _ _ _ (N.BBTemplate [Err _]) _ _) -> 1
+          _ -> 0
+
       (IsActiveEnable n) ->
         let (e, ty, _) = bbInputs b !! n in
-        case (e, ty) of
-          (Literal Nothing (BoolLit True), Bool)  -> 0
-          -- TODO: Emit warning? If enable signal is inferred as always False,
-          -- TODO: the component will never be enabled. This is probably not the
-          -- TODO: user's intention.
-          (Literal Nothing (BoolLit False), Bool) -> 1
-          (_, Bool)                               -> 1
+        case ty of
+          Enable _ ->
+            case e of
+              DataCon _ _ [Literal Nothing (BoolLit True)]  -> 0
+              -- TODO: Emit warning? If enable signal is inferred as always
+              -- TODO: False, the component will never be enabled. This is
+              -- TODO: probably not the user's intention.
+              DataCon _ _ [Literal Nothing (BoolLit False)] -> 1
+              _                                             -> 1
+          Bool ->
+            case e of
+              Literal Nothing (BoolLit True)  -> 0
+              -- TODO: Emit warning? If enable signal is inferred as always
+              -- TODO: False, the component will never be enabled. This is
+              -- TODO: probably not the user's intention.
+              Literal Nothing (BoolLit False) -> 1
+              _                               -> 1
           _ ->
             error $ $(curLoc) ++ "IsActiveEnable: Expected Bool, not: " ++ show ty
 
@@ -494,10 +521,10 @@ renderElem b (IF c t f) = do
                 | t1 == Text.pack t2 -> 1
                 | otherwise -> 0
               Nothing -> error $ $(curLoc) ++ "Expected a string literal: " ++ show e
-      (And es)   -> if all (/=0) (map (check iw syn) es)
+      (And es)   -> if all (/=0) (map (check xOpt iw syn) es)
                        then 1
                        else 0
-      CmpLE e1 e2 -> if check iw syn e1 <= check iw syn e2
+      CmpLE e1 e2 -> if check xOpt iw syn e1 <= check xOpt iw syn e2
                         then 1
                         else 0
       _ -> error $ $(curLoc) ++ "IF: condition must be: SIZE, LENGTH, IW64, LIT, ISLIT, or ISARG"
@@ -511,9 +538,16 @@ parseFail t = case runParse t of
   Success templ -> templ
 
 idToExpr
-  :: (Text,HWType)
-  -> (Expr,HWType,Bool)
-idToExpr (t,ty) = (Identifier (Text.toStrict t) Nothing,ty,False)
+  :: (Text, HWType)
+  -> (Expr, HWType, Bool)
+idToExpr (t, ty) =
+  (Identifier (Id.unsafeMake (Text.toStrict t)) Nothing, ty, False)
+
+bbResult :: HasCallStack => String -> BlackBoxContext -> (Expr, HWType)
+bbResult _s (bbResults -> [r]) = r
+bbResult s ctx = error [I.i|
+  Multi result primitives not supported when using template tag #{s}. Tag used
+  in blackbox implementation of #{bbName ctx} |]
 
 -- | Fill out the template corresponding to an output/input assignment of a
 -- component instantiation, and turn it into a single identifier so it can
@@ -530,7 +564,7 @@ lineToIdentifier b = foldrM (\e a -> do
 lineToType :: BlackBoxContext
            -> BlackBoxTemplate
            -> HWType
-lineToType b [(Typ Nothing)]  = snd $ bbResult b
+lineToType b [(Typ Nothing)]  = snd $ bbResult "~TYPO" b
 lineToType b [(Typ (Just n))] = let (_,ty,_) = bbInputs b !! n
                                 in  ty
 lineToType b [(TypElem t)]    = case lineToType b [t] of
@@ -550,13 +584,11 @@ renderTag :: Backend backend
           -> Element
           -> State backend Text
 renderTag _ (Text t)        = return t
-renderTag b (Result esc)    = do
-  escape <- if esc then unextend else pure id
-  fmap (Text.fromStrict . escape . Text.toStrict . renderOneLine) . getMon . expr False . fst $ bbResult b
-renderTag b (Arg esc n)  = do
+renderTag b (Result)    = do
+  fmap renderOneLine . getMon . expr False . fst $ bbResult "~RESULT" b
+renderTag b (Arg n)  = do
   let (e,_,_) = bbInputs b !! n
-  escape <- if esc then unextend else pure id
-  (Text.fromStrict . escape . Text.toStrict . renderOneLine) <$> getMon (expr False e)
+  renderOneLine <$> getMon (expr False e)
 
 renderTag b (Const n)  = do
   let (e,_,_) = bbInputs b !! n
@@ -602,13 +634,13 @@ renderTag b (Sel e n) =
   let ty = lineToType b [e]
   in  renderOneLine <$> getMon (hdlRecSel ty n)
 
-renderTag b (Typ Nothing)   = fmap renderOneLine . getMon . hdlType Internal . snd $ bbResult b
+renderTag b (Typ Nothing)   = fmap renderOneLine . getMon . hdlType Internal . snd $ bbResult "~TYPO" b
 renderTag b (Typ (Just n))  = let (_,ty,_) = bbInputs b !! n
                               in  renderOneLine <$> getMon (hdlType Internal ty)
-renderTag b (TypM Nothing)  = fmap renderOneLine . getMon . hdlTypeMark . snd $ bbResult b
+renderTag b (TypM Nothing)  = fmap renderOneLine . getMon . hdlTypeMark . snd $ bbResult "~TYPMO" b
 renderTag b (TypM (Just n)) = let (_,ty,_) = bbInputs b !! n
                               in  renderOneLine <$> getMon (hdlTypeMark ty)
-renderTag b (Err Nothing)   = fmap renderOneLine . getMon . hdlTypeErrValue . snd $ bbResult b
+renderTag b (Err Nothing)   = fmap renderOneLine . getMon . hdlTypeErrValue . snd $ bbResult "~ERRORO" b
 renderTag b (Err (Just n))  = let (_,ty,_) = bbInputs b !! n
                               in  renderOneLine <$> getMon (hdlTypeErrValue ty)
 renderTag b (Size e)        = return . Text.pack . show . typeSize $ lineToType b [e]
@@ -708,12 +740,12 @@ renderTag b (Template filenameL sourceL) = case file of
           source   <- elementsToText b sourceL
           return (Text.unpack filename, Text.unpack source)
 
-renderTag b CompName = pure (Text.fromStrict (bbCompName b))
+renderTag b CompName = pure (Id.toLazyText (bbCompName b))
 
 renderTag b CtxName = case bbCtxName b of
   Just nm -> return (Text.fromStrict nm)
-  _ | Identifier t _ <- fst (bbResult b)
-    -> return (Text.fromStrict t)
+  _ | Identifier t _ <- fst (bbResult "~CTXNAME" b)
+    -> return (Id.toLazyText t)
   _ -> error "internal error"
 
 
@@ -763,11 +795,13 @@ exprToString
 exprToString (Literal _ (NumLit i)) = Just (show i)
 exprToString (Literal _ (StringLit l)) = Just l
 exprToString (BlackBoxE "Clash.Promoted.Symbol.SSymbol" _ _ _ _ ctx _) =
-  let (e',_,_) = head (bbInputs ctx)
-  in  exprToString e'
+  case bbInputs ctx of
+    (e0,_,_):_ -> exprToString e0
+    _ -> error "internal error: insufficient bbInputs"
 exprToString (BlackBoxE "GHC.CString.unpackCString#" _ _ _ _ ctx _) =
-  let (e',_,_) = head (bbInputs ctx)
-  in  exprToString e'
+  case bbInputs ctx of
+    (e0,_,_):_ -> exprToString e0
+    _ -> error "internal error: insufficient bbInputs"
 exprToString _ = Nothing
 
 prettyBlackBox :: Monad m
@@ -782,15 +816,18 @@ prettyElem
 prettyElem (Text t) = return t
 prettyElem (Component (Decl i 0 args)) = do
   args' <- mapM (\(a,b) -> (,) <$> prettyBlackBox a <*> prettyBlackBox b) args
-  renderOneLine <$>
-    (nest 2 (string "~INST" <+> int i <> line <>
-        string "~OUTPUT" <+> string "=>" <+> string (fst (head args')) <+> string (snd (head args')) <+> string "~" <> line <>
-        vcat (mapM (\(a,b) -> string "~INPUT" <+> string "=>" <+> string a <+> string b <+> string "~") (tail args')))
-      <> line <> string "~INST")
+  case args' of
+    (arg:rest) ->
+      renderOneLine <$>
+        (nest 2 (string "~INST" <+> int i <> line <>
+            string "~OUTPUT" <+> string "=>" <+> string (fst arg) <+> string (snd arg) <+> string "~" <> line <>
+            vcat (mapM (\(a,b) -> string "~INPUT" <+> string "=>" <+> string a <+> string b <+> string "~") rest))
+          <> line <> string "~INST")
+    _ -> error "internal error: insufficient args"
 prettyElem (Component (Decl {})) =
   error $ $(curLoc) ++ "prettyElem can't (yet) render ~INST when subfuncion /= 0!"
-prettyElem (Result b) = if b then return "~ERESULT" else return "~RESULT"
-prettyElem (Arg b i) = renderOneLine <$> (if b then string "~EARG" else string "~ARG" <> brackets (int i))
+prettyElem Result = return "~RESULT"
+prettyElem (Arg i) = renderOneLine <$> ("~ARG" <> brackets (int i))
 prettyElem (Lit i) = renderOneLine <$> (string "~LIT" <> brackets (int i))
 prettyElem (Const i) = renderOneLine <$> (string "~CONST" <> brackets (int i))
 prettyElem (Name i) = renderOneLine <$> (string "~NAME" <> brackets (int i))
@@ -864,6 +901,7 @@ prettyElem (IsLit i) = renderOneLine <$> (string "~ISLIT" <> brackets (int i))
 prettyElem (IsVar i) = renderOneLine <$> (string "~ISVAR" <> brackets (int i))
 prettyElem (IsActiveHigh i) = renderOneLine <$> (string "~ISACTIVEHIGH" <> brackets (int i))
 prettyElem (IsActiveEnable i) = renderOneLine <$> (string "~ISACTIVEENABLE" <> brackets (int i))
+prettyElem (IsUndefined i) = renderOneLine <$> (string "~ISUNDEFINED" <> brackets (int i))
 
 -- Domain attributes:
 prettyElem (Tag i) = renderOneLine <$> (string "~TAG" <> brackets (int i))
@@ -940,8 +978,8 @@ walkElement f el = maybeToList (f el) ++ walked
         GenSym es _ -> concatMap go es
         DevNull es -> concatMap go es
         Text _ -> []
-        Result _ -> []
-        Arg _ _ -> []
+        Result -> []
+        Arg _ -> []
         ArgGen _ _ -> []
         Const _ -> []
         Lit _ -> []
@@ -973,6 +1011,7 @@ walkElement f el = maybeToList (f el) ++ walked
         IsInitDefined _ -> []
         IsActiveHigh _ -> []
         IsActiveEnable _ -> []
+        IsUndefined _ -> []
         StrCmp es _ -> concatMap go es
         OutputWireReg _ -> []
         Vars _ -> []
@@ -982,17 +1021,18 @@ walkElement f el = maybeToList (f el) ++ walked
 
 -- | Determine variables used in an expression. Used for VHDL sensitivity list.
 -- Also see: https://github.com/clash-lang/clash-compiler/issues/365
-usedVariables :: Expr -> [Identifier]
+usedVariables :: Expr -> [N.IdentifierText]
 usedVariables Noop              = []
-usedVariables (Identifier i _)  = [i]
+usedVariables (Identifier i _)  = [Id.toText i]
 usedVariables (DataCon _ _ es)  = concatMap usedVariables es
-usedVariables (DataTag _ e')    = [either id id e']
+usedVariables (DataTag _ e')    = [Id.toText (either id id e')]
 usedVariables (Literal {})      = []
-usedVariables (ConvBV _ _ _ e') = usedVariables e'
+usedVariables (ToBv _ _ e') = usedVariables e'
+usedVariables (FromBv _ _ e') = usedVariables e'
 usedVariables (IfThenElse e1 e2 e3) = concatMap usedVariables [e1,e2,e3]
 usedVariables (BlackBoxE _ _ _ _ t bb _) = nub (sList ++ sList')
   where
-    matchArg (Arg _ i) = Just i
+    matchArg (Arg i) = Just i
     matchArg _         = Nothing
 
     matchVar (ToVar [Text v] _) = Just (Text.toStrict v)
@@ -1010,11 +1050,12 @@ getUsedArguments (N.BBTemplate t) = nub (concatMap (walkElement matchArg) t)
   where
     matchArg =
       \case
-        Arg _ i -> Just i
+        Arg i -> Just i
         Component (Decl i _ _) -> Just i
         Const i -> Just i
         IsLit i -> Just i
         IsActiveEnable i -> Just i
+        IsUndefined i -> Just i
         Lit i -> Just i
         Name i -> Just i
         ToVar _ i -> Just i
@@ -1050,7 +1091,7 @@ getUsedArguments (N.BBTemplate t) = nub (concatMap (walkElement matchArg) t)
         MaxIndex _ -> Nothing
         OutputWireReg _ -> Nothing
         Repeat _ _ -> Nothing
-        Result _ -> Nothing
+        Result -> Nothing
         Sel _ _ -> Nothing
         SigD _ _ -> Nothing
         Size _ -> Nothing
