@@ -30,13 +30,28 @@ import qualified Data.Map                         as Map
 import qualified Data.Maybe                       as Maybe
 import qualified Data.Set                         as Set
 import qualified Data.Set.Lens                    as Lens
-import           Data.Text.Prettyprint.Doc        (vcat)
 
+#if MIN_VERSION_prettyprinter(1,7,0)
+import           Prettyprinter                    (vcat)
+#else
+import           Data.Text.Prettyprint.Doc        (vcat)
+#endif
+
+#if MIN_VERSION_ghc(9,0,0)
+import           GHC.Types.Basic                  (InlineSpec (..))
+#else
 import           BasicTypes                       (InlineSpec (..))
+#endif
 
 import           Clash.Annotations.BitRepresentation.Internal
   (CustomReprs)
-import           Clash.Core.Evaluator.Types       (PrimStep, PrimUnwind)
+
+#if EXPERIMENTAL_EVALUATOR
+import           Clash.Core.PartialEval           (Evaluator)
+#else
+import           Clash.Core.Evaluator.Types       (Evaluator)
+#endif
+
 import           Clash.Core.FreeVars
   (freeLocalIds, globalIds, globalIdOccursIn, localIdDoesNotOccurIn)
 import           Clash.Core.Pretty                (PrettyOptions(..), showPpr, showPpr', ppr)
@@ -62,28 +77,24 @@ import           Clash.Netlist.Util
   (splitNormalized)
 import           Clash.Normalize.Strategy
 import           Clash.Normalize.Transformations
-  (appPropFast, bindConstantVar, caseCon, flattenLet, reduceConst, topLet,
-   reduceNonRepPrim, removeUnusedExpr, deadCode)
 import           Clash.Normalize.Types
 import           Clash.Normalize.Util
 import           Clash.Primitives.Types           (CompiledPrimMap)
 import           Clash.Rewrite.Combinators        ((>->),(!->),repeatR,topdownR)
 import           Clash.Rewrite.Types
-  (RewriteEnv (..), RewriteState (..), bindings, dbgLevel, extra,
+  (RewriteEnv (..), RewriteState (..), bindings, dbgLevel, dbgRewriteHistoryFile, extra,
    tcCache, topEntities)
 import           Clash.Rewrite.Util
   (apply, isUntranslatableType, runRewriteSession)
 import           Clash.Util
 import           Clash.Util.Interpolate           (i)
 
-#ifdef HISTORY
 import           Data.Binary                      (encode)
 import qualified Data.ByteString                  as BS
 import qualified Data.ByteString.Lazy             as BL
 
 import           System.IO.Unsafe                 (unsafePerformIO)
 import           Clash.Rewrite.Types (RewriteStep(..))
-#endif
 
 
 -- | Run a NormalizeSession in a given environment
@@ -102,8 +113,8 @@ runNormalization
   -- ^ TyCon cache
   -> IntMap TyConName
   -- ^ Tuple TyCon cache
-  -> (PrimStep, PrimUnwind)
-  -- ^ Hardcoded evaluator (delta-reduction)
+  -> Evaluator
+  -- ^ Hardcoded evaluator for partial evaluation
   -> CompiledPrimMap
   -- ^ Primitive Definitions
   -> VarEnv Bool
@@ -121,6 +132,7 @@ runNormalization opts supply globals typeTrans reprs tcm tupTcm eval primMap rcs
                   (opt_dbgTransformations opts)
                   (opt_dbgTransformationsFrom opts)
                   (opt_dbgTransformationsLimit opts)
+                  (opt_dbgRewriteHistoryFile opts)
                   (opt_aggressiveXOpt opts)
                   typeTrans
                   tcm
@@ -128,6 +140,7 @@ runNormalization opts supply globals typeTrans reprs tcm tupTcm eval primMap rcs
                   eval
                   (mkVarSet topEnts)
                   reprs
+                  (opt_evaluatorFuelLimit opts)
 
     rwState   = RewriteState
                   0
@@ -135,7 +148,12 @@ runNormalization opts supply globals typeTrans reprs tcm tupTcm eval primMap rcs
                   supply
                   (error $ $(curLoc) ++ "Report as bug: no curFun",noSrcSpan)
                   0
+#if EXPERIMENTAL_EVALUATOR
+                  IntMap.empty
+                  0
+#else
                   (IntMap.empty, 0)
+#endif
                   emptyVarEnv
                   normState
 
@@ -165,12 +183,12 @@ normalize top = do
   newNormalized <- normalize (concat new)
   return (unionVarEnv (mkVarEnv topNormalized) newNormalized)
 
-normalize' :: Id -> NormalizeSession ([Id], (Id, Binding))
+normalize' :: Id -> NormalizeSession ([Id], (Id, Binding Term))
 normalize' nm = do
   exprM <- lookupVarEnv nm <$> Lens.use bindings
   let nmS = showPpr (varName nm)
   case exprM of
-    Just (Binding nm' sp inl tm) -> do
+    Just (Binding nm' sp inl pr tm) -> do
       tcm <- Lens.view tcCache
       topEnts <- Lens.view topEntities
       let isTop = nm `elemVarSet` topEnts
@@ -197,7 +215,7 @@ normalize' nm = do
       resTyRep <- not <$> isUntranslatableType False resTy
       if resTyRep
          then do
-            tmNorm <- normalizeTopLvlBndr isTopEnt nm (Binding nm' sp inl tm)
+            tmNorm <- normalizeTopLvlBndr isTopEnt nm (Binding nm' sp inl pr tm)
             let usedBndrs = Lens.toListOf globalIds (bindingTerm tmNorm)
             traceIf (nm `elem` usedBndrs)
                     (concat [ $(curLoc),"Expr belonging to bndr: ",nmS ," (:: "
@@ -230,7 +248,7 @@ normalize' nm = do
                             , showPpr (varType nm')
                             , ") has a non-representable return type."
                             , " Not normalising:\n", showPpr tm] )
-                    (return ([],(nm,(Binding nm' sp inl tm))))
+                    (return ([],(nm,(Binding nm' sp inl pr tm))))
 
 
     Nothing -> error $ $(curLoc) ++ "Expr belonging to bndr: " ++ nmS ++ " not found"
@@ -248,7 +266,7 @@ checkNonRecursive norm = case mapMaybeVarEnv go norm of
                                  | (a,b) <- eltsVarEnv rcs
                                  ])
  where
-  go (Binding nm _ _ tm) =
+  go (Binding nm _ _ _ tm) =
     if nm `globalIdOccursIn` tm
        then Just (nm,tm)
        else Nothing
@@ -272,8 +290,8 @@ cleanupGraph _ norm = return norm
 -- additional bindings which are used. See "Clash.Driver.Types.Binding".
 --
 data CallTree
-  = CLeaf   (Id, Binding)
-  | CBranch (Id, Binding) [CallTree]
+  = CLeaf   (Id, Binding Term)
+  | CBranch (Id, Binding Term) [CallTree]
 
 mkCallTree
   :: [Id]
@@ -314,8 +332,8 @@ stripArgs _ _ _ = Nothing
 flattenNode
   :: CallTree
   -> NormalizeSession (Either CallTree ((Id,Term),[CallTree]))
-flattenNode c@(CLeaf (_,(Binding _ _ NoInline _))) = return (Left c)
-flattenNode c@(CLeaf (nm,(Binding _ _ _ e))) = do
+flattenNode c@(CLeaf (_,(Binding _ _ NoInline _ _))) = return (Left c)
+flattenNode c@(CLeaf (nm,(Binding _ _ _ _ e))) = do
   isTopEntity <- elemVarSet nm <$> Lens.view topEntities
   if isTopEntity then return (Left c) else do
     tcm  <- Lens.view tcCache
@@ -328,9 +346,9 @@ flattenNode c@(CLeaf (nm,(Binding _ _ _ e))) = do
                return (Right ((nm,mkApps (mkTicks fun ticks) (reverse remainder)),[]))
           _ -> return (Right ((nm,e),[]))
       _ -> return (Right ((nm,e),[]))
-flattenNode b@(CBranch (_,(Binding _ _ NoInline _)) _) =
+flattenNode b@(CBranch (_,(Binding _ _ NoInline _ _)) _) =
   return (Left b)
-flattenNode b@(CBranch (nm,(Binding _ _ _ e)) us) = do
+flattenNode b@(CBranch (nm,(Binding _ _ _ _ e)) us) = do
   isTopEntity <- elemVarSet nm <$> Lens.view topEntities
   if isTopEntity then return (Left b) else do
     tcm  <- Lens.view tcCache
@@ -352,7 +370,7 @@ flattenCallTree
   :: CallTree
   -> NormalizeSession CallTree
 flattenCallTree c@(CLeaf _) = return c
-flattenCallTree (CBranch (nm,(Binding nm' sp inl tm)) used) = do
+flattenCallTree (CBranch (nm,(Binding nm' sp inl pr tm)) used) = do
   flattenedUsed   <- mapM flattenCallTree used
   (newUsed,il_ct) <- partitionEithers <$> mapM flattenNode flattenedUsed
   let (toInline,il_used) = unzip il_ct
@@ -361,10 +379,12 @@ flattenCallTree (CBranch (nm,(Binding nm' sp inl tm)) used) = do
     [] -> return tm
     _  -> do
       let tm1 = substTm "flattenCallTree.flattenExpr" subst tm
-#ifdef HISTORY
-      -- NB: When HISTORY is on, emit binary data holding the recorded rewrite steps
-      let !_ = unsafePerformIO
-             $ BS.appendFile "history.dat"
+
+      -- NB: When -fclash-debug-history is on, emit binary data holding the recorded rewrite steps
+      rewriteHistFile <- Lens.view dbgRewriteHistoryFile
+      when (Maybe.isJust rewriteHistFile) $
+        let !_ = unsafePerformIO
+             $ BS.appendFile (Maybe.fromJust rewriteHistFile)
              $ BL.toStrict
              $ encode RewriteStep
                  { t_ctx    = []
@@ -373,7 +393,7 @@ flattenCallTree (CBranch (nm,(Binding nm' sp inl tm)) used) = do
                  , t_before = tm
                  , t_after  = tm1
                  }
-#endif
+        in pure ()
       rewriteExpr ("flattenExpr",flatten) (showPpr nm, tm1) (nm', sp)
   let allUsed = newUsed ++ concat il_used
   -- inline all components when the resulting expression after flattening
@@ -386,27 +406,31 @@ flattenCallTree (CBranch (nm,(Binding nm' sp inl tm)) used) = do
                                         (Maybe.catMaybes toInline')
         let tm1 = substTm "flattenCallTree.flattenCheap" subst' newExpr
         newExpr' <- rewriteExpr ("flattenCheap",flatten) (showPpr nm, tm1) (nm', sp)
-        return (CBranch (nm,(Binding nm' sp inl newExpr')) (concat allUsed'))
-     else return (CBranch (nm,(Binding nm' sp inl newExpr)) allUsed)
+        return (CBranch (nm,(Binding nm' sp inl pr newExpr')) (concat allUsed'))
+     else return (CBranch (nm,(Binding nm' sp inl pr newExpr)) allUsed)
   where
     flatten =
       repeatR (topdownR (apply "appPropFast" appPropFast >->
                  apply "bindConstantVar" bindConstantVar >->
                  apply "caseCon" caseCon >->
+#if EXPERIMENTAL_EVALUATOR
+                 apply "deadcode" deadCode >->
+#else
                  (apply "reduceConst" reduceConst !-> apply "deadcode" deadCode) >->
+#endif
                  apply "reduceNonRepPrim" reduceNonRepPrim >->
                  apply "removeUnusedExpr" removeUnusedExpr >->
                  apply "flattenLet" flattenLet)) !->
       topdownSucR (apply "topLet" topLet)
 
-    goCheap c@(CLeaf   (nm2,(Binding _ _ inl2 e)))
+    goCheap c@(CLeaf   (nm2,(Binding _ _ inl2 _ e)))
       | inl2 == NoInline = (Nothing     ,[c])
       | otherwise        = (Just (nm2,e),[])
-    goCheap c@(CBranch (nm2,(Binding _ _ inl2 e)) us)
+    goCheap c@(CBranch (nm2,(Binding _ _ inl2 _ e)) us)
       | inl2 == NoInline = (Nothing, [c])
       | otherwise        = (Just (nm2,e),us)
 
-callTreeToList :: [Id] -> CallTree -> ([Id], [(Id, Binding)])
+callTreeToList :: [Id] -> CallTree -> ([Id], [(Id, Binding Term)])
 callTreeToList visited (CLeaf (nm,bndr))
   | nm `elem` visited = (visited,[])
   | otherwise         = (nm:visited,[(nm,bndr)])
