@@ -26,6 +26,7 @@
   >       C -> h x
 -}
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecursiveDo #-}
 {-# LANGUAGE TemplateHaskell #-}
@@ -54,9 +55,31 @@ import qualified Data.Map.Strict                  as Map
 import qualified Data.Maybe                       as Maybe
 import           Data.Monoid                      (All (..))
 
+#if MIN_VERSION_ghc(8,10,0)
+import           GHC.Hs.Utils                     (chunkify,mkChunkified)
+#else
+import           HsUtils                          (chunkify,mkChunkified)
+#endif
+
+#if MIN_VERSION_ghc(9,0,0)
+import           GHC.Settings.Constants           (mAX_TUPLE_SIZE)
+#else
+import           Constants                        (mAX_TUPLE_SIZE)
+#endif
+
+#if EXPERIMENTAL_EVALUATOR
+import           System.IO.Unsafe
+#endif
+
 -- internal
-import Clash.Core.DataCon    (DataCon, dcTag)
-import Clash.Core.Evaluator  (whnf')
+import Clash.Core.DataCon    (DataCon)
+
+#if EXPERIMENTAL_EVALUATOR
+import Clash.Core.PartialEval
+#else
+import Clash.Core.Evaluator.Types  (whnf')
+#endif
+
 import Clash.Core.FreeVars
   (termFreeVars', typeFreeVars', localVarsDoNotOccurIn)
 import Clash.Core.Literal    (Literal (..))
@@ -64,17 +87,16 @@ import Clash.Core.Term
   (LetBinding, Pat (..), PrimInfo (..), Term (..), TickInfo (..), collectArgs,
    collectArgsTicks, mkApps, mkTicks, patIds)
 import Clash.Core.TermInfo   (termType)
-import Clash.Core.TyCon      (tyConDataCons)
+import Clash.Core.TyCon      (TyConMap, TyConName, tyConDataCons)
 import Clash.Core.Type       (Type, isPolyFunTy, mkTyConApp, splitFunForallTy)
-import Clash.Core.Util       (sccLetBindings)
-import Clash.Core.Var        (isGlobalId)
+import Clash.Core.Util       (mkInternalVar, mkSelectorCase, sccLetBindings)
+import Clash.Core.Var        (isGlobalId, isLocalId)
 import Clash.Core.VarEnv
   (InScopeSet, elemInScopeSet, extendInScopeSetList, notElemInScopeSet, unionInScope)
 import Clash.Normalize.Types (NormalizeState)
 import Clash.Rewrite.Types
-  (RewriteMonad, bindings, evaluator, globalHeap, tcCache, tupleTcCache, uniqSupply)
-import Clash.Rewrite.Util    (mkInternalVar, mkSelectorCase,
-                              isUntranslatableType, isConstant)
+import Clash.Rewrite.Util    (isUntranslatableType)
+import Clash.Rewrite.WorkFree (isConstant)
 import Clash.Unique          (lookupUniqMap)
 import Clash.Util
 
@@ -145,13 +167,22 @@ collectGlobals' is0 substitution seen e@(collectArgsTicks -> (fun, args@(_:_), t
   | not eIsconstant = do
     tcm <- Lens.view tcCache
     bndrs <- Lens.use bindings
-    (primEval, primUnwind) <- Lens.view evaluator
-    gh <- Lens.use globalHeap
+    evaluate <- Lens.view evaluator
     ids <- Lens.use uniqSupply
     let (ids1,ids2) = splitSupply ids
     uniqSupply Lens..= ids2
-    let eval = (Lens.view Lens._3) . whnf' primEval primUnwind bndrs tcm gh ids1 is0 False
-        eTy  = termType tcm e
+#if EXPERIMENTAL_EVALUATOR
+    (i,_) <- Lens.use curFun
+    heap <- Lens.use ioHeap
+    addr <- Lens.use ioAddr
+    fuel <- Lens.view fuelLimit
+    let genv = mkGlobalEnv bndrs tcm is0 ids1 fuel heap addr
+    let eval = fst . unsafePerformIO . nf evaluate genv False i
+#else
+    gh <- Lens.use globalHeap
+    let eval = (Lens.view Lens._3) . whnf' evaluate bndrs tcm gh ids1 is0 False
+#endif
+    let eTy  = termType tcm e
     untran <- isUntranslatableType False eTy
     case untran of
       -- Don't lift out non-representable values, because they cannot be let-bound
@@ -369,29 +400,25 @@ disJointSelProj
   -> RewriteMonad NormalizeState (Maybe LetBinding,[Term])
 disJointSelProj _ _ (Leaf []) = return (Nothing,[])
 disJointSelProj inScope argTys cs = do
+    tcm    <- Lens.view tcCache
+    tupTcm <- Lens.view tupleTcCache
     let maxIndex = length argTys - 1
         css = map (\i -> fmap ((:[]) . (!!i)) cs) [0..maxIndex]
     (untran,tran) <- List.partitionM (isUntranslatableType False . snd) (zip [0..] argTys)
     let untranCs   = map (css!!) (map fst untran)
-        untranSels = zipWith (\(_,ty) cs' -> genCase ty Nothing []  cs')
+        untranSels = zipWith (\(_,ty) cs' -> genCase tcm tupTcm ty [ty]  cs')
                              untran untranCs
     (lbM,projs) <- case tran of
       []       -> return (Nothing,[])
-      [(i,ty)] -> return (Nothing,[genCase ty Nothing [] (css!!i)])
+      [(i,ty)] -> return (Nothing,[genCase tcm tupTcm ty [ty] (css!!i)])
       tys      -> do
-        tcm    <- Lens.view tcCache
-        tupTcm <- Lens.view tupleTcCache
         let m            = length tys
-            Just tupTcNm = IM.lookup m tupTcm
-            Just tupTc   = lookupUniqMap tupTcNm tcm
-            [tupDc]      = tyConDataCons tupTc
             (tyIxs,tys') = unzip tys
-            tupTy        = mkTyConApp tupTcNm tys'
+            tupTy        = mkBigTupTy tcm tupTcm tys'
             cs'          = fmap (\es -> map (es !!) tyIxs) cs
-            djCase       = genCase tupTy (Just tupDc) tys' cs'
+            djCase       = genCase tcm tupTcm tupTy tys' cs'
         scrutId <- mkInternalVar inScope "tupIn" tupTy
-        projections <- mapM (mkSelectorCase ($(curLoc) ++ "disJointSelProj")
-                                            inScope tcm (Var scrutId) (dcTag tupDc)) [0..m-1]
+        projections <- mapM (mkBigTupSelector inScope tcm tupTcm (Var scrutId) tys') [0..m-1]
         return (Just (scrutId,djCase),projections)
     let selProjs = tranOrUnTran 0 (zip (map fst untran) untranSels) projs
 
@@ -417,7 +444,7 @@ areShared inScope xs@(x:_) = noFV1 && allEqual xs
     Left tm  -> getAll (Lens.foldMapOf (termFreeVars' isLocallyBound)
                                        (const (All False)) tm)
 
-  isLocallyBound v = v `notElemInScopeSet` inScope
+  isLocallyBound v = isLocalId v && v `notElemInScopeSet` inScope
 
 -- | Create a list of arguments given a map of positions to common arguments,
 -- and a list of arguments
@@ -433,17 +460,16 @@ mkDJArgs n ((m,x):cms) (y:uncms)
 
 -- | Create a case-expression that selects between the distinct arguments given
 -- a case-tree
-genCase :: Type -- ^ Type of the alternatives
-        -> Maybe DataCon -- ^ DataCon to pack multiple arguments
+genCase :: TyConMap
+        -> IM.IntMap TyConName
+        -> Type -- ^ Type of the alternatives
         -> [Type] -- ^ Types of the arguments
         -> CaseTree [Term] -- ^ CaseTree of arguments
         -> Term
-genCase ty dcM argTys = go
+genCase tcm tupTcm ty argTys = go
   where
     go (Leaf tms) =
-      case dcM of
-        Just dc -> mkApps (Data dc) (map Right argTys ++ map Left tms)
-        _ -> head tms
+      mkBigTupTm tcm tupTcm (List.zipEqual argTys tms)
 
     go (LB lb ct) =
       Letrec lb (go ct)
@@ -458,6 +484,67 @@ genCase ty dcM argTys = go
     go (Branch scrut pats) =
       Case scrut ty (map (second go) pats)
 
+-- | Lookup the TyConName and DataCon for a tuple of size n
+findTup :: TyConMap -> IM.IntMap TyConName -> Int -> (TyConName,DataCon)
+findTup tcm tupTcm n = (tupTcNm,tupDc)
+  where
+    tupTcNm      = Maybe.fromMaybe (error $ $curLoc ++ "Can't find " ++ show n ++ "-tuple") $ IM.lookup n tupTcm
+    Just tupTc   = lookupUniqMap tupTcNm tcm
+    [tupDc]      = tyConDataCons tupTc
+
+mkBigTupTm :: TyConMap -> IM.IntMap TyConName -> [(Type,Term)] -> Term
+mkBigTupTm tcm tupTcm args = snd $ mkBigTup tcm tupTcm args
+
+mkSmallTup,mkBigTup :: TyConMap -> IM.IntMap TyConName -> [(Type,Term)] -> (Type,Term)
+mkSmallTup _ _ [] = error $ $curLoc ++ "mkSmallTup: Can't create 0-tuple"
+mkSmallTup _ _ [(ty,tm)] = (ty,tm)
+mkSmallTup tcm tupTcm args = (ty,tm)
+  where
+    (argTys,tms) = unzip args
+    (tupTcNm,tupDc) = findTup tcm tupTcm (length args)
+    tm = mkApps (Data tupDc) (map Right argTys ++ map Left tms)
+    ty = mkTyConApp tupTcNm argTys
+
+mkBigTup tcm tupTcm = mkChunkified (mkSmallTup tcm tupTcm)
+
+mkSmallTupTy,mkBigTupTy
+  :: TyConMap
+  -> IM.IntMap TyConName
+  -> [Type]
+  -> Type
+mkSmallTupTy _ _ [] = error $ $curLoc ++ "mkSmallTupTy: Can't create 0-tuple"
+mkSmallTupTy _ _ [ty] = ty
+mkSmallTupTy tcm tupTcm tys = mkTyConApp tupTcNm tys
+  where
+    m = length tys
+    (tupTcNm,_) = findTup tcm tupTcm m
+
+mkBigTupTy tcm tupTcm = mkChunkified (mkSmallTupTy tcm tupTcm)
+
+mkSmallTupSelector,mkBigTupSelector
+  :: MonadUnique m
+  => InScopeSet
+  -> TyConMap
+  -> IM.IntMap TyConName
+  -> Term
+  -> [Type]
+  -> Int
+  -> m Term
+mkSmallTupSelector _ _ _ scrut [_] 0 = return scrut
+mkSmallTupSelector _ _ _ _     [_] n = error $ $curLoc ++ "mkSmallTupSelector called with one type, but to select " ++ show n
+mkSmallTupSelector inScope tcm _ scrut _ n = mkSelectorCase ($curLoc ++ "mkSmallTupSelector") inScope tcm scrut 1 n
+
+mkBigTupSelector inScope tcm tupTcm scrut tys n = go (chunkify tys)
+  where
+    go [_] = mkSmallTupSelector inScope tcm tupTcm scrut tys n
+    go tyss = do
+      let (nOuter,nInner) = divMod n mAX_TUPLE_SIZE
+          tyss' = map (mkSmallTupTy tcm tupTcm) tyss
+      outer <- mkSmallTupSelector inScope tcm tupTcm scrut tyss' nOuter
+      inner <- mkSmallTupSelector inScope tcm tupTcm outer (tyss List.!! nOuter) nInner
+      return inner
+
+
 -- | Determine if a term in a function position is interesting to lift out of
 -- of a case-expression.
 --
@@ -521,11 +608,19 @@ interestingToLift inScope eval e@(Prim pInfo) args ticks
       ,("GHC.Base.modInt",lastNotPow2)
       ,("GHC.Classes.divInt#",lastNotPow2)
       ,("GHC.Classes.modInt#",lastNotPow2)
+#if MIN_VERSION_base(4,15,0)
+      ,("GHC.Num.Integer.integerMul",allNonPow2)
+      ,("GHC.Num.Integer.integerDiv",lastNotPow2)
+      ,("GHC.Num.Integer.integerMod",lastNotPow2)
+      ,("GHC.Num.Integer.integerQuot",lastNotPow2)
+      ,("GHC.Num.Integer.integerRem",lastNotPow2)
+#else
       ,("GHC.Integer.Type.timesInteger",allNonPow2)
       ,("GHC.Integer.Type.divInteger",lastNotPow2)
       ,("GHC.Integer.Type.modInteger",lastNotPow2)
       ,("GHC.Integer.Type.quotInteger",lastNotPow2)
       ,("GHC.Integer.Type.remInteger",lastNotPow2)
+#endif
       ,("GHC.Prim.*#",allNonPow2)
       ,("GHC.Prim.quotInt#",lastNotPow2)
       ,("GHC.Prim.remInt#",lastNotPow2)
