@@ -6,6 +6,7 @@
   Utility functions used by the normalisation transformations
 -}
 
+{-# LANGUAGE CPP #-}
 {-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE RecordWildCards #-}
@@ -46,8 +47,13 @@ import qualified Data.HashSet            as HashSet
 import           Data.Text               (Text)
 import qualified Data.Text as Text
 
+#if MIN_VERSION_ghc(9,0,0)
+import           GHC.Builtin.Names       (eqTyConKey)
+import           GHC.Types.Unique        (getKey)
+#else
 import           PrelNames               (eqTyConKey)
 import           Unique                  (getKey)
+#endif
 
 import           Clash.Annotations.Primitive (extractPrim)
 import           Clash.Core.FreeVars
@@ -137,11 +143,12 @@ alreadyInlined f cf = do
     Nothing       -> return Nothing
     Just inlined' -> return (lookupVarEnv f inlined')
 
+-- | Record a new inlining in the `inlineHistory`
 addNewInline
   :: Id
-  -- ^ Function we want to inline
+  -- ^ Function we're inlining
   -> Id
-  -- ^ Function in which we want to perform the inlining
+  -- ^ Function in which we're inlining it
   -> NormalizeMonad ()
 addNewInline f cf =
   inlineHistory %= extendVarEnvWith
@@ -226,7 +233,7 @@ mergeCsrs
   -> [TickInfo]
   -- ^ Ticks to wrap around proposed new term
   -> Term
-  -- ^ "Old" term
+  -- ^ \"Old\" term
   -> ([Either Term Type] -> Term)
   -- ^ Proposed new term in case any constants were found
   -> [Either Term Type]
@@ -408,9 +415,9 @@ isCheapFunction tm = case classifyFunction tm of
 normalizeTopLvlBndr
   :: Bool
   -> Id
-  -> Binding
-  -> NormalizeSession Binding
-normalizeTopLvlBndr isTop nm (Binding nm' sp inl tm) = makeCachedU nm (extra.normalized) $ do
+  -> Binding Term
+  -> NormalizeSession (Binding Term)
+normalizeTopLvlBndr isTop nm (Binding nm' sp inl pr tm) = makeCachedU nm (extra.normalized) $ do
   tcm <- Lens.view tcCache
   let nmS = showPpr (varName nm)
   -- We deshadow the term because sometimes GHC gives us
@@ -424,7 +431,7 @@ normalizeTopLvlBndr isTop nm (Binding nm' sp inl tm) = makeCachedU nm (extra.nor
   tm3 <- rewriteExpr ("normalization",normalization) (nmS,tm2) (nm',sp)
   curFun .= old
   let ty' = termType tcm tm3
-  return (Binding nm'{varType = ty'} sp inl tm3)
+  return (Binding nm'{varType = ty'} sp inl pr tm3)
 
 -- | Turn type equality constraints into substitutions and apply them.
 --
@@ -517,7 +524,8 @@ removedTm
   :: Type
   -> Term
 removedTm =
-  TyApp (Prim (PrimInfo "Clash.Transformations.removedArg" undefinedTy WorkNever))
+  let removedNm = "Clash.Transformations.removedArg" in
+  TyApp (Prim (PrimInfo removedNm undefinedTy WorkNever SingleResult))
 
 -- | A tick to prefix an inlined expression with it's original name.
 -- For example, given
