@@ -13,6 +13,7 @@ instantiations.
 {-# LANGUAGE OverloadedStrings #-}
 {-# LANGUAGE PatternSynonyms   #-}
 {-# LANGUAGE QuasiQuotes       #-}
+{-# LANGUAGE RecordWildCards   #-}
 {-# LANGUAGE TemplateHaskell   #-}
 {-# LANGUAGE TypeApplications  #-}
 {-# LANGUAGE ViewPatterns      #-}
@@ -20,13 +21,16 @@ instantiations.
 module Clash.Primitives.DSL
   (
   -- * Annotations
-    blackBoxHaskell
+    BlackBoxHaskellOpts(..)
+  , blackBoxHaskell
 
   -- * declarations
   , BlockState (..)
   , TExpr
   , declaration
+  , declarationReturn
   , instDecl
+  , instHO
   , viaAnnotatedSignal
 
   -- ** Literals
@@ -34,16 +38,18 @@ module Clash.Primitives.DSL
   , LitHDL (..)
   , pattern High
   , pattern Low
+  , constructProduct
   , tuple
   , vec
 
   -- ** Extraction
   , tInputs
-  , tResult
+  , tResults
   , getStr
   , getBool
   , exprToInteger
   , tExprToInteger
+  , deconstructProduct
   , untuple
   , unvec
 
@@ -72,7 +78,11 @@ module Clash.Primitives.DSL
 import           Clash.Util                      (HasCallStack, clogBase)
 import           Control.Lens                    hiding (Indexed, assign)
 import           Control.Monad.State
+import           Data.Default                    (Default(def))
+import           Data.IntMap                     (IntMap)
+import qualified Data.IntMap                     as IntMap
 import           Data.List                       (intersperse)
+import           Data.List.Extra                 (zipEqual)
 import           Data.Maybe                      (fromMaybe)
 import           Data.Semigroup                  hiding (Product)
 import           Data.Semigroup.Monad
@@ -86,57 +96,116 @@ import           Clash.Annotations.Primitive     (HDL (..), Primitive (..))
 import           Clash.Backend                   hiding (fromBV, toBV)
 import           Clash.Backend.VHDL              (VHDLState)
 import           Clash.Core.Var                  (Attr')
-import           Clash.Netlist.BlackBox.Util     (exprToString)
-import           Clash.Netlist.Id
-import           Clash.Netlist.Types             hiding (toBit)
-import           Clash.Netlist.Util              hiding (mkUniqueIdentifier)
+import           Clash.Netlist.BlackBox.Util     (exprToString, renderElem)
+import           Clash.Netlist.BlackBox.Types
+  (BlackBoxTemplate, Element(Component, Text), Decl(..))
+import qualified Clash.Netlist.Id                as Id
+import           Clash.Netlist.Types             hiding (Component, toBit)
+import           Clash.Netlist.Util
 import qualified Data.String.Interpolate         as I
 import           Data.String.Interpolate.Util    (unindent)
 import           Language.Haskell.TH             (Name)
 import           Prelude
 
+-- | Options for 'blackBoxHaskell' function. Use 'def' from package
+-- 'data-default' for a set of default options.
+data BlackBoxHaskellOpts = BlackBoxHaskellOpts
+  { -- | Arguments to ignore (i.e., remove during normalization)
+    --
+    -- Default: []
+    bo_ignoredArguments :: [Int]
+
+    -- | HDLs to use the blackbox for
+    --
+    -- Default: all
+  , bo_supportedHdls :: [HDL]
+
+    -- | Does this blackbox assign its results to multiple binders?
+    --
+    -- Default: False.
+  , bo_multiResult :: Bool
+  }
+
+instance Default BlackBoxHaskellOpts where
+  def = BlackBoxHaskellOpts
+    { bo_ignoredArguments = []
+    , bo_supportedHdls = [minBound..maxBound]
+    , bo_multiResult = False
+    }
+
 -- | Create a blackBoxHaskell primitive. To be used as part of an annotation:
 --
 -- @
--- {-# ANN myFunction (blackBoxHaskell [1,2,3] VHDL 'myFunction 'myBBF) #-}
+-- {-# ANN myFunction (blackBoxHaskell 'myFunction 'myBBF def{_ignoredArguments=[2,3]}) #-}
 -- @
+--
+-- [2,3] would mean this blackbox __ignores__ its second and third argument.
 blackBoxHaskell
-  :: [Int]
-  -- ^ Ignored arguments
-  -> HDL
-  -- ^ hdl the blackbox is for
-  -> Name
+  :: Name
   -- ^ blackbox name
   -> Name
   -- ^ template function name
+  -> BlackBoxHaskellOpts
+  -- ^ Options, see data structure for more information
   -> Primitive
-blackBoxHaskell (show -> ign) hdl bb tf =
-  InlinePrimitive [hdl] $ unindent [I.i|
+blackBoxHaskell bb tf BlackBoxHaskellOpts{..} =
+  InlinePrimitive bo_supportedHdls $ unindent [I.i|
   [ { "BlackBoxHaskell" :
-       { "name" : "#{bb}"
-       , "templateFunction" : "#{tf}"
-       , "ignoredArguments" : #{ign}
-       }
+      { "name" : "#{bb}"
+      , "templateFunction" : "#{tf}"
+      , "ignoredArguments" : #{show bo_ignoredArguments}
+      , "multiResult" : #{toJsonBool bo_multiResult}
+      }
     }
-  ]
-|]
+  ] |]
+ where
+  toJsonBool :: Bool -> String
+  toJsonBool True = "true"
+  toJsonBool False = "false"
 
 -- | The state of a block. Contains a list of declarations and a the
 --   backend state.
 data BlockState backend = BlockState
   { _bsDeclarations :: [Declaration]
-  , _bsBackend      :: backend
+    -- ^ Declarations store
+  , _bsHigherOrderCalls :: IntMap Int
+    -- ^ Tracks how many times a higher order function has been instantiated.
+    -- Needed to fill in the second field of 'Clash.Netlist.BlackBox.Types.Decl'
+  , _bsBackend :: backend
+    -- ^ Backend state
   }
+makeLenses ''BlockState
+
+instance Backend backend => HasIdentifierSet (BlockState backend) where
+  identifierSet :: Lens' (BlockState backend) IdentifierSet
+  identifierSet = bsBackend . identifierSet
 
 -- | A typed expression.
 data TExpr = TExpr
   { ety :: HWType
   , eex :: Expr
   } deriving Show
-
-makeLenses ''BlockState
 makeLenses ''TExpr
 
+-- | Run a block declaration. Assign the result of the block builder to the
+-- result variable in the given blackbox context.
+declarationReturn
+  :: Backend backend
+  => BlackBoxContext
+  -> Text.Text
+  -- ^ block name
+  -> State (BlockState backend) [TExpr]
+  -- ^ block builder yielding an expression that should be assigned to the
+  -- result variable in the blackbox context
+  -> State backend Doc
+  -- ^ pretty printed block
+declarationReturn bbCtx blockName blockBuilder =
+  declaration blockName $ do
+    res <- blockBuilder
+    forM_ (zip (bbResults bbCtx) res) $ \(rNm, r) -> do
+      let (Identifier resultNm Nothing, _) = rNm
+      addDeclaration (Assignment resultNm (eex r))
+
 -- | Run a block declaration.
 declaration
   :: Backend backend
@@ -148,25 +217,45 @@ declaration
   -- ^ pretty printed block
 declaration blockName s = do
   backend0 <- get
-  let initState = BlockState [] backend0
-      BlockState decs backend1 = execState s initState
+  let initState = BlockState [] IntMap.empty backend0
+      BlockState decs _hoCalls backend1 = execState s initState
   put backend1
-  blockNameUnique <- mkUniqueIdentifier Basic blockName
+  blockNameUnique <- Id.makeBasic blockName
   getMon $ blockDecl blockNameUnique (reverse decs)
 
 -- | Add a declaration to the state.
 addDeclaration :: Declaration -> State (BlockState backend) ()
 addDeclaration dec = bsDeclarations %= cons dec
 
--- | Create a new basic unique name using the given Text as a template.
-newName :: Backend backend => Text -> State (BlockState backend) Identifier
-newName nm = zoom bsBackend $ mkUniqueIdentifier Basic nm
+-- | Declare a new signal with the given name and type.
+declare'
+  :: Backend backend
+  => Text
+  -- ^ Name hint
+  -> WireOrReg
+  -- ^ Should signal be declared as a wire or a reg
+  -> HWType
+  -- ^ Type of new signal
+  -> State (BlockState backend) Identifier
+  -- ^ Expression pointing the the new signal
+declare' decName wireOrReg ty = do
+  uniqueName <- Id.makeBasic decName
+  addDeclaration (NetDecl' Nothing wireOrReg uniqueName (Right ty) Nothing)
+  pure uniqueName
 
 -- | Declare a new signal with the given name and type.
-declare :: Backend backend => Identifier -> HWType -> State (BlockState backend) TExpr
-declare decName ty = do
-  uniqueName <- newName decName
-  addDeclaration (NetDecl Nothing uniqueName ty)
+declare
+  :: Backend backend
+  => Text
+  -- ^ Name hint
+  -> WireOrReg
+  -- ^ Should signal be declared as a wire or a reg
+  -> HWType
+  -- ^ Type of new signal
+  -> State (BlockState backend) TExpr
+  -- ^ Expression pointing the the new signal
+declare decName wireOrReg ty = do
+  uniqueName <- declare' decName wireOrReg ty
   pure (TExpr ty (Identifier uniqueName Nothing))
 
 -- | Assign an expression to an identifier, returns the new typed
@@ -180,7 +269,7 @@ assign
   -> State (BlockState backend) TExpr
   -- ^ the identifier of the expression that actually got assigned
 assign aName (TExpr ty aExpr) = do
-  texp@(~(TExpr _ (Identifier uniqueName Nothing))) <- declare aName ty
+  texp@(~(TExpr _ (Identifier uniqueName Nothing))) <- declare aName Wire ty
   addDeclaration (Assignment uniqueName aExpr)
   pure texp
 
@@ -189,7 +278,7 @@ assign aName (TExpr ty aExpr) = do
 -- will be used to assign the given expression to which is subsequently indexed.
 unvec
   :: Backend backend
-  => Identifier
+  => Text
   -- ^ Name hint for intermediate signal
   -> TExpr
   -- ^ Vector expression
@@ -201,21 +290,34 @@ unvec vName v@(ety -> Vector vSize eType) = do
   pure (map (TExpr eType . vIndex) [0..vSize-1])
 unvec _ e = error $ "unvec: cannot be called on non-vector: " <> show (ety e)
 
+-- | Extract the fields of a product type and return expressions
+--   to them. These new expressions are given unique names and get
+--   declared in the block scope.
+deconstructProduct
+  :: (HasCallStack, Backend backend)
+  => TExpr
+  -- ^ Product expression
+  -> [Text]
+  -- ^ Name hints for element assignments
+  -> State (BlockState backend) [TExpr]
+deconstructProduct (TExpr ty@(Product _ _ tys) (Identifier resName _)) vals = do
+  newNames <- zipWithM (flip declare Wire) vals tys
+  addDeclaration $ Assignment resName $ DataCon ty (DC (ty, 0)) (fmap eex newNames)
+  pure newNames
+deconstructProduct e i =
+  error $ "deconstructProduct: " <> show e <> " " <> show i
+
 -- | Extract the elements of a tuple expression and return expressions
 --   to them. These new expressions are given unique names and get
---   declared the block scope.
+--   declared in the block scope.
 untuple
   :: (HasCallStack, Backend backend)
   => TExpr
   -- ^ Tuple expression
-  -> [Identifier]
+  -> [Text]
   -- ^ Name hints for element assignments
   -> State (BlockState backend) [TExpr]
-untuple (TExpr ty@(Product _ _ tys) (Identifier resName _)) vals = do
-  newNames <- zipWithM declare vals tys
-  addDeclaration $ Assignment resName $ DataCon ty (DC (ty, 0)) (fmap eex newNames)
-  pure newNames
-untuple e i = error $ "untuple: " <> show e <> " " <> show i
+untuple = deconstructProduct
 
 -- | The high literal bit.
 pattern High :: TExpr
@@ -252,7 +354,7 @@ bvLit sz n =
 -- | Convert a bool to a bit.
 boolToBit
   :: (HasCallStack, Backend backend)
-  => Identifier
+  => Text
   -- ^ Name hint for intermediate signal
   -> TExpr
   -> State (BlockState backend) TExpr
@@ -260,7 +362,7 @@ boolToBit bitName = \case
   T -> pure High
   F -> pure Low
   TExpr Bool boolExpr -> do
-    texp@(~(TExpr _ (Identifier uniqueBitName Nothing))) <- declare bitName Bit
+    texp@(~(TExpr _ (Identifier uniqueBitName Nothing))) <- declare bitName Wire Bit
     addDeclaration $
       CondAssignment uniqueBitName Bit boolExpr Bool
         [ (Just (BoolLit True), Literal Nothing (BitLit H))
@@ -274,22 +376,22 @@ boolToBit bitName = \case
 --   Returns a reference to a declared `Bit` that should get assigned
 --   by something (usually the output port of an entity).
 boolFromBit
-  :: Backend backend
-  => Identifier
+  :: Text
   -- ^ Name hint for intermediate signal
   -> TExpr
-  -> State (BlockState backend) TExpr
+  -> State (BlockState VHDLState) TExpr
 boolFromBit = outputCoerce Bit Bool (<> " = '1'")
 
 -- | Used to create an output `Bool` from a `BitVector` of given size.
 -- Works in a similar way to `boolFromBit` above.
+--
+-- TODO: Implement for (System)Verilog
 boolFromBitVector
-  :: Backend backend
-  => Size
-  -> Identifier
+  :: Size
+  -> Text
   -- ^ Name hint for intermediate signal
   -> TExpr
-  -> State (BlockState backend) TExpr
+  -> State (BlockState VHDLState) TExpr
 boolFromBitVector n =
   outputCoerce (BitVector n) Bool (\i -> "unsigned(" <> i <> ") > 0")
 
@@ -299,7 +401,7 @@ boolFromBitVector n =
 -- TODO: Implement for (System)Verilog
 unsignedFromBitVector
   :: Size
-  -> Identifier
+  -> Text
   -- ^ Name hint for intermediate signal
   -> TExpr
   -> State (BlockState VHDLState) TExpr
@@ -314,7 +416,7 @@ unsignedFromBitVector n =
 --
 -- TODO: Implement for (System)Verilog
 boolFromBits
-  :: [Identifier]
+  :: [Text]
   -> TExpr
   -> State (BlockState VHDLState) [TExpr]
 boolFromBits inNames = outputFn (map (const Bit) inNames) Bool
@@ -329,18 +431,19 @@ outputCoerce
   :: (HasCallStack, Backend backend)
   => HWType
   -> HWType
-  -> (Identifier -> Identifier)
-  -> Identifier
+  -> (Text -> Text)
+  -> Text
   -> TExpr
   -> State (BlockState backend) TExpr
-outputCoerce fromType toType exprStringFn inName expr_
+outputCoerce fromType toType exprStringFn inName0 expr_
   | TExpr outType (Identifier outName Nothing) <- expr_
   , outType == toType = do
-      inName' <- newName inName
-      let exprIdent = Identifier (exprStringFn inName') Nothing
-      addDeclaration (NetDecl Nothing inName' fromType)
+      inName1 <- Id.makeBasic inName0
+      let inName2 = Id.unsafeMake (exprStringFn (Id.toText inName1))
+          exprIdent = Identifier inName2 Nothing
+      addDeclaration (NetDecl Nothing inName1 fromType)
       addDeclaration (Assignment outName exprIdent)
-      pure (TExpr fromType (Identifier inName' Nothing))
+      pure (TExpr fromType (Identifier inName1 Nothing))
 outputCoerce _ toType _ _ texpr = error $ "outputCoerce: the expression " <> show texpr
                                   <> " must be an Identifier with type " <> show toType
 
@@ -354,19 +457,20 @@ outputFn
   :: (HasCallStack, Backend backend)
   => [HWType]
   -> HWType
-  -> ([Identifier] -> Identifier)
-  -> [Identifier]
+  -> ([Text] -> Text)
+  -> [Text]
   -> TExpr
   -> State (BlockState backend) [TExpr]
-outputFn fromTypes toType exprFn inNames (TExpr outType (Identifier outName Nothing))
+outputFn fromTypes toType exprFn inNames0 (TExpr outType (Identifier outName Nothing))
   | outType == toType = do
-      inNames' <- mapM newName inNames
-      let exprIdent = Identifier (exprFn inNames') Nothing
+      inNames1 <- mapM Id.makeBasic inNames0
+      let idExpr = Id.unsafeMake (exprFn (map Id.toText inNames1))
+          exprIdent = Identifier idExpr Nothing
       sequenceOf_ each [ addDeclaration (NetDecl Nothing nm t)
-                       | (nm,t) <- zip inNames' fromTypes ]
+                       | (nm, t) <- zip inNames1 fromTypes ]
       addDeclaration (Assignment outName exprIdent)
       pure [ TExpr t (Identifier nm Nothing)
-           | (nm,t) <- zip inNames' fromTypes ]
+           | (nm,t) <- zipEqual inNames1 fromTypes ]
 outputFn _ outType _ _ texpr =
   error $ "outputFn: the expression " <> show texpr
   <> " must be an Identifier with type " <> show outType
@@ -387,14 +491,18 @@ vec els@(el:_)
   theVec = mkVectorChain (length els) (ety el) (map eex els)
 vec [] = error "vec: can't be used on empty lists"
 
+-- | Construct a product type given its type and fields
+constructProduct :: HWType -> [TExpr] -> TExpr
+constructProduct ty els =
+  TExpr ty (DataCon ty (DC (ty,0)) (map eex els))
+
 -- | Create an n-tuple of 'TExpr'
 tuple :: [TExpr] -> TExpr
 tuple [] = error $ "nTuple: Cannot create empty tuple"
 tuple [_] =
   -- If we don't put this in: tuple . untuple /= id
   error $ "nTuple: Cannot create 1-tuple"
-tuple els =
-  TExpr tupTy (DataCon tupTy (DC (tupTy,0)) (map eex els))
+tuple els = constructProduct tupTy els
  where
   commas = Text.replicate (length els - 1) ","
   tupTy = Product ("GHC.Tuple.(" <> commas <> ")") Nothing (map ety els)
@@ -421,8 +529,8 @@ exprToInteger _ = Nothing
 --   bitvector if the expression is not already a bitvector.
 toBV
   :: Backend backend
-  => Identifier
-  -- ^ BitVector name
+  => Text
+  -- ^ BitVector name hint
   -> TExpr
   -- ^ expression
   -> State (BlockState backend) TExpr
@@ -430,28 +538,28 @@ toBV
 toBV bvName a = case a of
   TExpr BitVector{} _ -> pure a
   TExpr aTy aExpr     -> assign bvName $
-    TExpr (BitVector (typeSize aTy)) (ConvBV Nothing aTy True aExpr)
+    TExpr (BitVector (typeSize aTy)) (ToBv Nothing aTy aExpr)
 
 -- | Assign an output bitvector to an expression. Declares a new
 --   bitvector if the expression is not already a bitvector.
 fromBV
   :: (HasCallStack, Backend backend)
-  => Identifier
-  -- ^ BitVector name
+  => Text
+  -- ^ BitVector name hint
   -> TExpr
   -- ^ expression
   -> State (BlockState backend) TExpr
   -- ^ bv expression
 fromBV _ a@(TExpr BitVector{} _) = pure a
 fromBV bvName (TExpr aTy (Identifier aName Nothing)) = do
-  bvName' <- newName bvName
-  let bvExpr = ConvBV Nothing aTy False (Identifier bvName' Nothing)
+  bvName' <- Id.makeBasic bvName
+  let bvExpr = FromBv Nothing aTy (Identifier bvName' Nothing)
       bvTy   = BitVector (typeSize aTy)
   addDeclaration (NetDecl Nothing bvName' bvTy)
   addDeclaration (Assignment aName bvExpr)
   pure (TExpr bvTy (Identifier bvName' Nothing))
 fromBV _ texpr = error $
-  "fromBV: the expression " <> show texpr <> "must be an Indentifier"
+  "fromBV: the expression " <> show texpr <> "must be an Identifier"
 
 clog2 :: Num i => Integer -> i
 clog2 = fromIntegral . fromMaybe 0 . clogBase 2
@@ -478,9 +586,75 @@ instance Num LitHDL where
 instance IsString LitHDL where
   fromString = S
 
+-- | Instantiate/call a higher-order function.
+instHO
+  :: Backend backend
+  => BlackBoxContext
+  -- ^ BlackBoxContext, used for rendering higher-order function and error
+  -- reporting
+  -> Int
+  -- ^ Position of HO-argument. For example:
+  --
+  --   fold :: forall n a . (a -> a -> a) -> Vec (n + 1) a -> a
+  --
+  -- would have its HO-argument at position 0, while
+  --
+  --  iterateI :: forall n a. KnownNat n => (a -> a) -> a -> Vec n a
+  --
+  -- would have it at position 1.
+  -> (HWType, BlackBoxTemplate)
+  -- ^ Result type of HO function
+  -> [(TExpr, BlackBoxTemplate)]
+  -- ^ Arguments and their types
+  -> State (BlockState backend) TExpr
+  -- ^ Result of the function
+instHO bbCtx fPos (resTy, bbResTy) argsWithTypes = do
+  let (args0, argTypes) = unzip argsWithTypes
+  fSubPos <- fromMaybe 0 . IntMap.lookup fPos <$> use bsHigherOrderCalls
+  bsHigherOrderCalls %= IntMap.insert fPos (succ fSubPos)
+
+  -- Create argument identifiers, example: fold_ho3_0_arg0
+  let
+    ctxName = last (Text.split (=='.') (bbName bbCtx))
+    baseArgName = ctxName <> "_" <> "ho" <> showt fPos <> "_" <> showt fSubPos
+    argName n = baseArgName <> "_arg" <> showt n
+  args1 <- zipWithM (\argN -> toIdentifier' (argName argN)) [(0::Int)..] args0
+
+  let
+    args2 = map (pure . Text . Id.toLazyText) args1
+
+    -- Create result identifier
+    -- See https://github.com/clash-lang/clash-compiler/issues/919 for info on
+    -- logic of 'resWireOrReg'
+    resWireOrReg =
+      case IntMap.lookup fPos (bbFunctions bbCtx) of
+        Just ((_,rw,_,_,_,_):_) -> rw
+        _ -> error "internal error"
+  resName <- declare' (ctxName <> "_" <> "ho" <> showt fPos <> "_"
+                               <> showt fSubPos <> "_res") resWireOrReg resTy
+  let res = ([Text (Id.toLazyText resName)], bbResTy)
+
+  -- Render HO argument to plain text
+  let component = Component (Decl fPos fSubPos (res:zip args2 argTypes))
+  rendered0 <-
+    zoom bsBackend (string =<< (renderElem bbCtx component <*> pure 0))
+
+  let
+    layout = LayoutOptions (AvailablePerLine 120 0.4)
+    rendered1 = renderLazy (layoutPretty layout rendered0)
+
+  addDeclaration $
+    BlackBoxD
+      ("__INST_" <> bbName bbCtx <> "_BB_INTERNAL__") [] [] []
+      (BBTemplate [Text rendered1])
+      (emptyBBContext ("__INST_" <> bbName bbCtx <> "_BB_INTERNAL__"))
+
+  pure (TExpr resTy (Identifier resName Nothing))
+
 -- | Instantiate a component/entity in a block state.
 instDecl
-  :: Backend backend
+  :: forall backend
+   . Backend backend
   => EntityOrComponent
   -- ^ Type of instantiation
   -> Identifier
@@ -499,15 +673,23 @@ instDecl entOrComp compName instLbl attrs inPorts outPorts = do
   inPorts' <- mapM (mkPort In) inPorts
   outPorts' <- mapM (mkPort Out) outPorts
 
-  addDeclaration $ InstDecl entOrComp Nothing compName instLbl (mkAttrs attrs) (inPorts' ++ outPorts')
+  addDeclaration $
+    InstDecl
+      entOrComp Nothing [] compName instLbl (mkAttrs attrs)
+      (NamedPortMap (inPorts' ++ outPorts'))
     where
-    mkPort inOrOut (nm, pExpr) = do
-      TExpr ty pExpr' <- toIdentifier (nm <> "_port")  pExpr
-      pure (Identifier nm Nothing, inOrOut, ty, pExpr')
+    mkPort
+      :: PortDirection
+      -> (Text, TExpr)
+      -> StateT (BlockState backend) Identity (Expr, PortDirection, HWType, Expr)
+    mkPort inOrOut (nmText, pExpr) = do
+      TExpr ty pExpr' <- toIdentifier (nmText <> "_port")  pExpr
+      pure (Identifier (Id.unsafeMake nmText) Nothing, inOrOut, ty, pExpr')
 
     -- Convert a list of name attributes to the form clash wants
     mkAttrs :: [(Text.Text, LitHDL)] -> [(Expr, HWType, Expr)]
-    mkAttrs = map (\(s, ty) -> (Identifier s Nothing, hdlTy ty, litExpr ty))
+    mkAttrs = map (\(s, ty) -> ( Identifier (Id.unsafeMake s) Nothing
+                               , hdlTy ty, litExpr ty) )
 
     litExpr :: LitHDL -> Expr
     litExpr (B b) = Literal Nothing (BoolLit b)
@@ -525,7 +707,7 @@ instDecl entOrComp compName instLbl attrs inPorts outPorts = do
 -- annotated type, using the given attributes.
 viaAnnotatedSignal
   :: (HasCallStack, Backend backend)
-  => Text
+  => Identifier
   -- ^ Name given to signal
   -> TExpr
   -- ^ expression the signal is assigned to
@@ -551,14 +733,14 @@ tInputs :: BlackBoxContext -> [(TExpr, HWType)]
 tInputs = map (\(x, t, _) -> (TExpr t x, t)) . bbInputs
 
 -- | The TExp result of a blackbox context.
-tResult :: BlackBoxContext -> TExpr
-tResult = (\(x,t) -> TExpr t x) . bbResult
+tResults :: BlackBoxContext -> [TExpr]
+tResults = map (\(x,t) -> TExpr t x) . bbResults
 
 -- | Get an identifier to an expression, creating a new assignment if
 --   necessary.
 toIdentifier'
   :: Backend backend
-  => Identifier
+  => Text
   -- ^ desired new identifier name, will be made unique
   -> TExpr
   -- ^ expression to get identifier of
@@ -573,7 +755,7 @@ toIdentifier' nm texp = do
 --   necessary.
 toIdentifier
   :: Backend backend
-  => Identifier
+  => Text
   -- ^ desired new identifier name, will be made unique
   -> TExpr
   -- ^ expression to get identifier of
@@ -586,7 +768,7 @@ toIdentifier nm texp = do
 -- | And together @(&&)@ two expressions, assigning it to a new identifier.
 andExpr
   :: Backend backend
-  => Identifier
+  => Text
   -- ^ name hint
   -> TExpr
   -- ^ a
@@ -599,8 +781,8 @@ andExpr _ F _     = pure F
 andExpr _ aExpr T = pure aExpr
 andExpr _ _ F     = pure F
 andExpr nm a b = do
-  aIdent <- toIdentifier' (nm <> "_a") a
-  bIdent <- toIdentifier' (nm <> "_b") b
+  aIdent <- Id.toText <$> toIdentifier' (nm <> "_a") a
+  bIdent <- Id.toText <$> toIdentifier' (nm <> "_b") b
   -- This is somewhat hacky and relies on the fact that clash doesn't
   -- postprocess the text in Identifier. The alternative is to run
   -- this as a fully fledged @BlackBoxE@ but that involves a lot of
@@ -611,12 +793,12 @@ andExpr nm a b = do
       VHDL          -> aIdent <> " and " <> bIdent
       Verilog       -> aIdent <> " && " <> bIdent
       SystemVerilog -> aIdent <> " && " <> bIdent
-  assign nm $ TExpr Bool (Identifier andTxt Nothing)
+  assign nm $ TExpr Bool (Identifier (Id.unsafeMake andTxt) Nothing)
 
 -- | Negate @(not)@ an expression, assigning it to a new identifier.
 notExpr
   :: Backend backend
-  => Identifier
+  => Text
   -- ^ name hint
   -> TExpr
   -- ^ a
@@ -625,13 +807,13 @@ notExpr
 notExpr _ T = pure F
 notExpr _ F = pure T
 notExpr nm aExpr = do
-  aIdent <- toIdentifier' (nm <> "_a") aExpr
+  aIdent <- Id.toText <$> toIdentifier' (nm <> "_a") aExpr
   -- See disclaimer in `andExpr` above.
   notTxt <- uses bsBackend hdlKind <&> \case
     VHDL          -> "not " <> aIdent
     Verilog       -> "! " <> aIdent
     SystemVerilog -> "! " <> aIdent
-  assign nm $ TExpr Bit (Identifier notTxt Nothing)
+  assign nm $ TExpr Bit (Identifier (Id.unsafeMake notTxt) Nothing)
 
 -- | Creates a BV that produces the following vhdl:
 --
@@ -641,7 +823,7 @@ notExpr nm aExpr = do
 --
 -- TODO: Implement for (System)Verilog
 pureToBV
-  :: Identifier
+  :: Text
   -- ^ name hint
   -> Int
   -- ^ Size (n)
@@ -650,10 +832,10 @@ pureToBV
   -> State (BlockState VHDLState) TExpr
   -- ^ (0 to n => ARG)
 pureToBV nm n arg = do
-  arg' <- toIdentifier' nm arg
+  arg' <- Id.toText <$> toIdentifier' nm arg
   -- This is very hard coded and hacky
   let text = "(0 to " <> showt n <> " => " <> arg' <> ")"
-  assign nm $ TExpr (BitVector (n+1)) (Identifier text Nothing)
+  assign nm $ TExpr (BitVector (n+1)) (Identifier (Id.unsafeMake text) Nothing)
 
 -- | Creates a BV that produces the following vhdl:
 --
@@ -663,7 +845,7 @@ pureToBV nm n arg = do
 --
 -- TODO: Implement for (System)Verilog
 pureToBVResized
-  :: Identifier
+  :: Text
   -- ^ name hint
   -> Int
   -- ^ Size (n)
@@ -672,14 +854,14 @@ pureToBVResized
   -> State (BlockState VHDLState) TExpr
   -- ^ std_logic_vector(resize(ARG, Size))
 pureToBVResized nm n arg = do
-  arg' <- toIdentifier' nm arg
+  arg' <- Id.toText <$> toIdentifier' nm arg
   -- This is very hard coded and hacky
   let text = "std_logic_vector(resize(" <> arg' <> ", " <> showt n <> "))"
-  assign nm $ TExpr (BitVector n) (Identifier text Nothing)
+  assign nm $ TExpr (BitVector n) (Identifier (Id.unsafeMake text) Nothing)
 
 -- | Allows assignment of a port to be "open"
 open
   :: Backend backend
   => HWType
   -> State (BlockState backend) TExpr
-open hwType = pure $ TExpr hwType (Identifier "open" Nothing)
+open hwType = pure $ TExpr hwType (Identifier (Id.unsafeMake "open") Nothing)
