@@ -18,9 +18,11 @@
 {-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE TemplateHaskell #-}
 
-module Clash.Rewrite.Util where
+module Clash.Rewrite.Util
+  ( module Clash.Rewrite.Util
+  , module Clash.Rewrite.WorkFree
+  ) where
 
-import           Control.Monad.Extra         (andM, eitherM)
 import           Control.Concurrent.Supply   (splitSupply)
 import           Control.DeepSeq
 import           Control.Exception           (throw)
@@ -40,10 +42,9 @@ import           Data.Functor.Const          (Const (..))
 import           Data.List                   (group, partition, sort)
 import qualified Data.List                   as List
 import qualified Data.List.Extra             as List
-import           Data.List.Extra             (allM, partitionM)
+import           Data.List.Extra             (partitionM)
 import qualified Data.Map                    as Map
 import           Data.Maybe
-  (catMaybes, isJust, mapMaybe, fromMaybe)
 import qualified Data.Monoid                 as Monoid
 import qualified Data.Set                    as Set
 import qualified Data.Set.Lens               as Lens
@@ -52,47 +53,49 @@ import qualified Data.Set.Ordered.Extra      as OSet
 import           Data.Text                   (Text)
 import qualified Data.Text                   as Text
 
-#ifdef HISTORY
+import           System.IO.Unsafe            (unsafePerformIO)
+
 import           Data.Binary                 (encode)
 import qualified Data.ByteString             as BS
 import qualified Data.ByteString.Lazy        as BL
-import           System.IO.Unsafe            (unsafePerformIO)
+
+#if MIN_VERSION_ghc(9,0,0)
+import           GHC.Types.Basic             (InlineSpec (..))
+#else
+import           BasicTypes                  (InlineSpec (..))
 #endif
 
-import           BasicTypes                  (InlineSpec (..))
+#if EXPERIMENTAL_EVALUATOR
+import           Clash.Core.PartialEval
+import           Clash.Core.PartialEval.NormalForm
+#else
+import           Clash.Core.Evaluator.Types  (PureHeap, whnf')
+#endif
 
-import           Clash.Core.DataCon          (dcExtTyVars)
-import           Clash.Core.Evaluator        (whnf')
-import           Clash.Core.Evaluator.Types  (PureHeap)
 import           Clash.Core.FreeVars
-  (freeLocalVars, hasLocalFreeVars, localIdDoesNotOccurIn, localIdOccursIn,
-   typeFreeVars, termFreeVars', freeLocalIds, globalIdOccursIn)
+  (freeLocalVars, localIdDoesNotOccurIn, localIdOccursIn,
+   typeFreeVars, termFreeVars', freeLocalIds)
 import           Clash.Core.Name
 import           Clash.Core.Pretty           (showPpr)
 import           Clash.Core.Subst
   (substTmEnv, aeqTerm, aeqType, extendIdSubst, mkSubst, substTm)
 import           Clash.Core.Term
 import           Clash.Core.TermInfo
-import           Clash.Core.TyCon
-  (TyConMap, tyConDataCons)
-import           Clash.Core.Type             (KindOrType, Type (..),
-                                              TypeView (..), coreView1,
-                                              normalizeType,
-                                              typeKind, tyView, isPolyFunTy)
-import           Clash.Core.Util
-  (dataConInstArgTysE, isClockOrReset, isEnable)
+import           Clash.Core.TyCon            (TyConMap)
+import           Clash.Core.Type             (Type (..), normalizeType, typeKind)
 import           Clash.Core.Var
-  (Id, IdScope (..), TyVar, Var (..), isLocalId, mkGlobalId, mkLocalId, mkTyVar)
+  (Id, IdScope (..), TyVar, Var (..), mkGlobalId, mkLocalId, mkTyVar)
 import           Clash.Core.VarEnv
   (InScopeSet, VarEnv, elemVarSet, extendInScopeSetList, mkInScopeSet,
    uniqAway, uniqAway', mapVarEnv, eltsVarEnv, unitVarSet, emptyVarEnv,
    mkVarEnv, eltsVarSet, elemVarEnv, lookupVarEnv, extendVarEnv)
 import           Clash.Debug
 import           Clash.Driver.Types
-  (DebugLevel (..), BindingMap, Binding(..))
+  (DebugLevel (..), BindingMap, Binding(..), IsPrim(..))
 import           Clash.Netlist.Util          (representableType)
 import           Clash.Pretty                (clashPretty, showDoc)
 import           Clash.Rewrite.Types
+import           Clash.Rewrite.WorkFree
 import           Clash.Unique
 import           Clash.Util
 import qualified Clash.Util.Interpolate as I
@@ -160,12 +163,13 @@ apply = \s rewrite ctx expr0 -> do
   let hasChanged = Monoid.getAny anyChanged
       !expr2     = if hasChanged then expr1 else expr0
   Monad.when hasChanged (transformCounter += 1)
-#ifdef HISTORY
-  -- NB: When HISTORY is on, emit binary data holding the recorded rewrite steps
-  Monad.when hasChanged $ do
+
+  -- NB: When -fclash-debug-history is on, emit binary data holding the recorded rewrite steps
+  rewriteHistFile <- Lens.view dbgRewriteHistoryFile
+  Monad.when (isJust rewriteHistFile && hasChanged) $ do
     (curBndr, _) <- Lens.use curFun
     let !_ = unsafePerformIO
-             $ BS.appendFile "history.dat"
+             $ BS.appendFile (fromJust rewriteHistFile)
              $ BL.toStrict
              $ encode RewriteStep
                  { t_ctx    = tfContext ctx
@@ -175,7 +179,6 @@ apply = \s rewrite ctx expr0 -> do
                  , t_after  = expr1
                  }
     return ()
-#endif
 
   dbgFrom <- Lens.view dbgTransformationsFrom
   dbgLimit <- Lens.view dbgTransformationsLimit
@@ -256,7 +259,7 @@ applyDebug lvl _transformations _fromLimit name exprOld hasChanged exprNew =
                      , "substitution."
                      ])
 
-    traceIf (lvl >= DebugApplied && (not (beforeTy `aeqType` afterTy)))
+    traceIf (lvl >= DebugApplied && (not (normalizeType tcm beforeTy `aeqType` normalizeType tcm afterTy)))
             ( concat [ $(curLoc)
                      , "Error when applying rewrite ", name
                      , " to:\n" , before
@@ -266,7 +269,7 @@ applyDebug lvl _transformations _fromLimit name exprOld hasChanged exprNew =
                      ]
             ) (return ())
 
-  Monad.when (lvl >= DebugApplied && not hasChanged && not (exprOld `aeqTerm` exprNew)) $
+  Monad.when (lvl >= DebugSilent && not hasChanged && not (exprOld `aeqTerm` exprNew)) $
     error $ $(curLoc) ++ "Expression changed without notice(" ++ name ++  "): before"
                       ++ before ++ "\nafter:\n" ++ after
 
@@ -298,7 +301,7 @@ runRewriteSession :: RewriteEnv
                   -> RewriteMonad extra a
                   -> a
 runRewriteSession r s m =
-  traceIf (_dbgLevel r > DebugNone)
+  traceIf (_dbgLevel r > DebugSilent)
     ("Clash: Applied " ++ show (s' ^. transformCounter) ++ " transformations")
     a
   where
@@ -355,19 +358,6 @@ mkBinderFor is tcm name (Right ty) = do
   let ki = typeKind tcm ty
   return (Right (mkTyVar ki (coerce name')))
 
--- | Make a new, unique, identifier
-mkInternalVar
-  :: (MonadUnique m)
-  => InScopeSet
-  -> OccName
-  -- ^ Name of the identifier
-  -> KindOrType
-  -> m Id
-mkInternalVar inScope name ty = do
-  i <- getUniqueM
-  let nm = mkUnsafeInternalName name i
-  return (uniqAway inScope (mkLocalId ty nm))
-
 -- | Inline the binders in a let-binding that have a certain property
 inlineBinders
   :: (Term -> LetBinding -> RewriteMonad extra Bool)
@@ -522,52 +512,6 @@ liftAndSubsituteBinders inScope toLift toKeep body = do
     else
       go subst2 inl
 
--- | Determines whether a global binder is work free. Errors if binder does
--- not exist.
-isWorkFreeBinder :: HasCallStack => Id -> RewriteMonad extra Bool
-isWorkFreeBinder bndr =
-  makeCachedU bndr workFreeBinders $ do
-    bExprM <- lookupVarEnv bndr <$> Lens.use bindings
-    case bExprM of
-      Nothing -> error ("isWorkFreeBinder: couldn't find binder: " ++ showPpr bndr)
-      Just (bindingTerm -> t) ->
-        if bndr `globalIdOccursIn` t
-        then pure False
-        else isWorkFree t
-
--- | Determine whether a term does any work, i.e. adds to the size of the circuit
-isWorkFree
-  :: Term
-  -> RewriteMonad extra Bool
-isWorkFree (collectArgs -> (fun,args)) = case fun of
-  Var i ->
-    if | isPolyFunTy (varType i) -> pure False
-       | isLocalId i -> pure True
-       | otherwise -> andM [isWorkFreeBinder i, allM isWorkFreeArg args]
-  Data {} -> allM isWorkFreeArg args
-  Literal {} -> pure True
-  Prim pInfo -> case primWorkInfo pInfo of
-    -- We can ignore the arguments, because this primitive outputs a constant
-    -- regardless of its arguments
-    WorkConstant -> pure True
-    WorkNever -> allM isWorkFreeArg args
-    WorkVariable -> pure (all isConstantArg args)
-    -- Things like clock or reset generator always perform work
-    WorkAlways -> pure False
-  Lam _ e -> andM [isWorkFree e, allM isWorkFreeArg args]
-  TyLam _ e -> andM [isWorkFree e, allM isWorkFreeArg args]
-  Letrec bs e ->
-    andM [isWorkFree e, allM (isWorkFree . snd) bs, allM isWorkFreeArg args]
-  Case s _ [(_,a)] ->
-    andM [isWorkFree s, isWorkFree a, allM isWorkFreeArg args]
-  Cast e _ _ ->
-    andM [isWorkFree e, allM isWorkFreeArg args]
-  _ ->
-    pure False
- where
-  isWorkFreeArg e = eitherM isWorkFree (pure . const True) (pure e)
-  isConstantArg = either isConstant (const True)
-
 isFromInt :: Text -> Bool
 isFromInt nm = nm == "Clash.Sized.Internal.BitVector.fromInteger##" ||
                nm == "Clash.Sized.Internal.BitVector.fromInteger#" ||
@@ -575,78 +519,6 @@ isFromInt nm = nm == "Clash.Sized.Internal.BitVector.fromInteger##" ||
                nm == "Clash.Sized.Internal.Signed.fromInteger#" ||
                nm == "Clash.Sized.Internal.Unsigned.fromInteger#"
 
--- | Determine if a term represents a constant
-isConstant :: Term -> Bool
-isConstant e = case collectArgs e of
-  (Data _, args)   -> all (either isConstant (const True)) args
-  (Prim _, args) -> all (either isConstant (const True)) args
-  (Lam _ _, _)     -> not (hasLocalFreeVars e)
-  (Literal _,_)    -> True
-  _                -> False
-
-isConstantNotClockReset
-  :: Term
-  -> RewriteMonad extra Bool
-isConstantNotClockReset e = do
-  tcm <- Lens.view tcCache
-  let eTy = termType tcm e
-  if isClockOrReset tcm eTy
-     then case collectArgs e of
-        (Prim p,_) -> return (primName p == "Clash.Transformations.removedArg")
-        _ -> return False
-     else pure (isConstant e)
-
--- TODO: Remove function after using WorkInfo in 'isWorkFreeIsh'
-isWorkFreeClockOrResetOrEnable
-  :: TyConMap
-  -> Term
-  -> Maybe Bool
-isWorkFreeClockOrResetOrEnable tcm e =
-  let eTy = termType tcm e in
-  if isClockOrReset tcm eTy || isEnable tcm eTy then
-    case collectArgs e of
-      (Prim p,_) -> Just (primName p == "Clash.Transformations.removedArg")
-      (Var _, []) -> Just True
-      (Data _, []) -> Just True -- For Enable True/False
-      (Literal _,_) -> Just True
-      _ -> Just False
-  else
-    Nothing
-
--- | A conservative version of 'isWorkFree'. Is used to determine in 'bindConstantVar'
--- to determine whether an expression can be "bound" (locally inlined). While
--- binding workfree expressions won't result in extra work for the circuit, it
--- might very well cause extra work for Clash. In fact, using 'isWorkFree' in
--- 'bindConstantVar' makes Clash two orders of magnitude slower for some of our
--- test cases.
---
--- In effect, this function is a version of 'isConstant' that also considers
--- references to clocks and resets constant. This allows us to bind
--- HiddenClock(ResetEnable) constructs, allowing Clash to constant spec
--- subconstants - most notably KnownDomain. Doing that enables Clash to
--- eliminate any case-constructs on it.
-isWorkFreeIsh
-  :: Term
-  -> RewriteMonad extra Bool
-isWorkFreeIsh e = do
-  tcm <- Lens.view tcCache
-  case isWorkFreeClockOrResetOrEnable tcm e of
-    Just b -> pure b
-    Nothing ->
-      case collectArgs e of
-        (Data _, args)   -> allM isWorkFreeIshArg args
-        (Prim pInfo, args) -> case primWorkInfo pInfo of
-          WorkAlways     -> pure False -- Things like clock or reset generator always
-                                       -- perform work
-          WorkVariable   -> pure (all isConstantArg args)
-          _              -> allM isWorkFreeIshArg args
-        (Lam _ _, _)     -> pure (not (hasLocalFreeVars e))
-        (Literal _,_)    -> pure True
-        _                -> pure False
- where
-  isWorkFreeIshArg = either isWorkFreeIsh (pure . const True)
-  isConstantArg    = either isConstant (const True)
-
 inlineOrLiftBinders
   :: (LetBinding -> RewriteMonad extra Bool)
   -- ^ Property test
@@ -746,6 +618,7 @@ liftBinding (var@Id {varName = idName} ,e) = do
 #else
                                       EmptyInlineSpec
 #endif
+                                      IsFun
                                       newBody)
              -- Return the new binder
              return (var, newExpr)
@@ -795,7 +668,7 @@ addGlobalBind
   -> RewriteMonad extra ()
 addGlobalBind vNm ty sp inl body = do
   let vId = mkGlobalId ty vNm
-  (ty,body) `deepseq` bindings %= extendUniqMap vNm (Binding vId sp inl body)
+  (ty,body) `deepseq` bindings %= extendUniqMap vNm (Binding vId sp inl IsFun body)
 
 -- | Create a new name out of the given name, but with another unique. Resulting
 -- unique is guaranteed to not be in the given InScopeSet.
@@ -848,49 +721,6 @@ isUntranslatableType stringRepresentable ty =
                              <*> Lens.view tcCache
                              <*> pure ty)
 
--- | Make a binder that should not be referenced
-mkWildValBinder
-  :: (MonadUnique m)
-  => InScopeSet
-  -> Type
-  -> m Id
-mkWildValBinder is = mkInternalVar is "wild"
-
--- | Make a case-decomposition that extracts a field out of a (Sum-of-)Product type
-mkSelectorCase
-  :: HasCallStack
-  => (Functor m, MonadUnique m)
-  => String -- ^ Name of the caller of this function
-  -> InScopeSet
-  -> TyConMap -- ^ TyCon cache
-  -> Term -- ^ Subject of the case-composition
-  -> Int -- n'th DataCon
-  -> Int -- n'th field
-  -> m Term
-mkSelectorCase caller inScope tcm scrut dcI fieldI = go (termType tcm scrut)
-  where
-    go (coreView1 tcm -> Just ty') = go ty'
-    go scrutTy@(tyView -> TyConApp tc args) =
-      case tyConDataCons (lookupUniqMap' tcm tc) of
-        [] -> cantCreate $(curLoc) ("TyCon has no DataCons: " ++ show tc ++ " " ++ showPpr tc) scrutTy
-        dcs | dcI > length dcs -> cantCreate $(curLoc) "DC index exceeds max" scrutTy
-            | otherwise -> do
-          let dc = indexNote ($(curLoc) ++ "No DC with tag: " ++ show (dcI-1)) dcs (dcI-1)
-          let (Just fieldTys) = dataConInstArgTysE inScope tcm dc args
-          if fieldI >= length fieldTys
-            then cantCreate $(curLoc) "Field index exceed max" scrutTy
-            else do
-              wildBndrs <- mapM (mkWildValBinder inScope) fieldTys
-              let ty = indexNote ($(curLoc) ++ "No DC field#: " ++ show fieldI) fieldTys fieldI
-              selBndr <- mkInternalVar inScope "sel" ty
-              let bndrs  = take fieldI wildBndrs ++ [selBndr] ++ drop (fieldI+1) wildBndrs
-                  pat    = DataPat dc (dcExtTyVars dc) bndrs
-                  retVal = Case scrut ty [ (pat, Var selBndr) ]
-              return retVal
-    go scrutTy = cantCreate $(curLoc) ("Type of subject is not a datatype: " ++ showPpr scrutTy) scrutTy
-
-    cantCreate loc info scrutTy = error $ loc ++ "Can't create selector " ++ show (caller,dcI,fieldI) ++ " for: (" ++ showPpr scrut ++ " :: " ++ showPpr scrutTy ++ ")\nAdditional info: " ++ info
-
 -- | Specialise an application on its argument
 specialise :: Lens' extra (Map.Map (Id, Int, Either Term Type) Id) -- ^ Lens into previous specialisations
            -> Lens' extra (VarEnv Int) -- ^ Lens into the specialisation history
@@ -953,7 +783,7 @@ specialise' specMapLbl specHistLbl specLimitLbl (TransformContext is0 _) e (Var
       -- Determine if we can specialize f
       bodyMaybe <- fmap (lookupUniqMap (varName f)) $ Lens.use bindings
       case bodyMaybe of
-        Just (Binding _ sp inl bodyTm) -> do
+        Just (Binding _ sp inl _ bodyTm) -> do
           -- Determine if we see a sequence of specialisations on a growing argument
           specHistM <- lookupUniqMap f <$> Lens.use (extra.specHistLbl)
           specLim   <- Lens.use (extra . specLimitLbl)
@@ -1022,6 +852,7 @@ specialise' specMapLbl specHistLbl specLimitLbl (TransformContext is0 _) e (Var
           bs' -> init bs' ++ bs
         go bs _ = bs
 
+-- Specialising non Var's is used by nonRepANF
 specialise' _ _ _ _ctx _ (appE,args,ticks) (Left specArg) = do
   -- Create binders and variable references for free variables in 'specArg'
   let (specBndrs,specVars) = specArgBndrsAndVars (Left specArg)
@@ -1121,17 +952,36 @@ whnfRW
 whnfRW isSubj ctx@(TransformContext is0 _) e rw = do
   tcm <- Lens.view tcCache
   bndrs <- Lens.use bindings
-  (primEval, primUnwind) <- Lens.view evaluator
+  eval <- Lens.view evaluator
   ids <- Lens.use uniqSupply
   let (ids1,ids2) = splitSupply ids
   uniqSupply Lens..= ids2
+
+#if EXPERIMENTAL_EVALUATOR
+  (i, _) <- Lens.use curFun
+  heap <- Lens.use ioHeap
+  addr <- Lens.use ioAddr
+  fuel <- Lens.view fuelLimit
+  let genv = mkGlobalEnv bndrs tcm is0 ids1 fuel heap addr
+
+  case unsafePerformIO (nf eval genv isSubj i e) of
+    (!e', !genv') -> do
+      ioHeap Lens..= genvHeap genv'
+      ioAddr Lens..= genvAddr genv'
+
+      -- TODO I remain unsure about this. Do I want to use bindPureHeap?
+      rw (ctx { tfInScope = genvInScope genv' }) e'
+#else
   gh <- Lens.use globalHeap
-  case whnf' primEval primUnwind bndrs tcm gh ids1 is0 isSubj e of
+
+  case whnf' eval bndrs tcm gh ids1 is0 isSubj e of
     (!gh1,ph,v) -> do
       globalHeap Lens..= gh1
       bindPureHeap tcm ph rw ctx v
+#endif
 {-# SCC whnfRW #-}
 
+#if !EXPERIMENTAL_EVALUATOR
 -- | Binds variables on the PureHeap over the result of the rewrite
 --
 -- To prevent unnecessary rewrites only do this when rewrite changed something.
@@ -1167,7 +1017,8 @@ bindPureHeap tcm heap rw ctx0@(TransformContext is0 hist) e = do
     --
     -- † https://github.com/clash-lang/clash-compiler/pull/1354#issuecomment-635430374
     -- ‡ https://www.microsoft.com/en-us/research/wp-content/uploads/2016/07/supercomp-by-eval.pdf
-    inlineBinders inlineTest ctx0 (Letrec bndrs e1) >>= \case
+    bs <- Lens.use bindings
+    inlineBinders (inlineTest bs) ctx0 (Letrec bndrs e1) >>= \case
       e2@(Letrec bnders1 e3) ->
         pure (fromMaybe e2 (removeUnusedBinders bnders1 e3))
       e2 ->
@@ -1175,23 +1026,20 @@ bindPureHeap tcm heap rw ctx0@(TransformContext is0 hist) e = do
   else
     return e1
   where
-    bndrs = map toLetBinding $ toListUniqMap heap
     heapIds = map fst bndrs
     is1 = extendInScopeSetList is0 heapIds
     ctx = TransformContext is1 (LetBody heapIds : hist)
 
+    bndrs = map toLetBinding $ toListUniqMap heap
+
     toLetBinding :: (Unique,Term) -> LetBinding
     toLetBinding (uniq,term) = (nm, term)
       where
         ty = termType tcm term
         nm = mkLocalId ty (mkUnsafeSystemName "x" uniq) -- See [Note: Name re-creation]
 
-    inlineTest _ (i, stripTicks -> e_) =
-      if isLocalVar e_ then
-        -- Don't inline `let x = x in x`, it throws  us in an infinite loop
-        pure (i `localIdDoesNotOccurIn` e_)
-      else
-        isWorkFree e_
+    inlineTest bs _ (_, stripTicks -> e_) = isWorkFree workFreeBinders bs e_
+#endif
 
 -- | Remove unused binders in given let-binding. Returns /Nothing/ if no unused
 -- binders were found.
